'Test of Monte Carlo optimizations for EMT'

from asap3 import EMT, MonteCarloEMT, MonteCarloAtoms, \
    Atoms, FullNeighborList, NeighborList
from ase.lattice.cubic import FaceCenteredCubic
from ase import data
from asap3.test.pytest_markers import ReportTest, serial

import numpy as np
import time
import sys

import pytest

def checkenergies(a, a2):
    err = 0
    if 0:
        print("    Checking Cartesian positions")
        err = abs((a.get_positions() - a2.get_positions()).ravel())
        idx = np.argmax(err)
        at, co = idx/3, idx%3
        print("      err =", err[idx], "at", idx, ":", at, co)
        ReportTest("      Worst Cartesian position (%d,%d)" % (at, co),
                   a.get_positions()[at,co],
                   a2.get_positions()[at,co], 1e-10)
    print("    Checking energies")
    e1 = a.get_potential_energies()
    e2 = a2.get_potential_energies()
    err = e1 - e2
    idx = np.argmax(err)
    ReportTest(f"      Worst energy ({idx})", e1[idx], e2[idx], 1e-5)

@serial
@pytest.mark.parametrize('pbctype', ('periodic', 'free', 'mixed'))
def test_montecarlo_energies(pbctype):
    element = "Cu"
    pbcdict = dict(periodic=True, free=False, mixed=(1,0,0))
    pbc = pbcdict[pbctype]

    magnarr = np.array((0.0, 0.01, 0.1, 1.0, 3.0, 10.0))
    numarr = np.array((1, 3, 10, 50, 200, 2000, -3))
    nrun = 1
    for ntest in range(nrun):
        print()
        print("Running pass", ntest+1, "of", nrun)
        print()
        print("Periodic boundaries:", pbc)
        if nrun > 1:
            time.sleep(1)
        atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(10,10,10),
                                symbol=element, pbc=pbc)
        atoms = MonteCarloAtoms(atoms)
        print("Number of atoms:", len(atoms))
        atoms.calc = MonteCarloEMT()
        atoms.get_potential_energies()  # Force creation of neighbor list.
        atoms2 = atoms.copy()
        atoms2.calc = EMT()

        print() 
        print("Testing perturbations of single atoms")

        pick1 = np.argsort(np.random.random((len(magnarr),)))
        for magn in np.take(magnarr, pick1):
            pick2 = np.argsort(np.random.random((len(numarr),)))
            for number in np.take(numarr, pick2):
                # Pick number random atoms.
                if number < 0:
                    # Find N neighboring atoms and perturb them
                    number = -number
                    nblist = atoms.calc.get_neighborlist()
                    n = 0
                    while len(nblist[n]) < number:
                        n += 1
                    pick = np.concatenate([np.array((n,)), nblist[n][:number-1]])
                    del nblist
                else:
                    pick = np.argsort(np.random.random((len(atoms),)))[:number]
                if number > 15:
                    s = f"<<< {number} atoms >>>"
                else:
                    s = str(list(pick))
                print("  dr = %.3f: %d atoms: %s" % (magn, number, s))
                for i in pick:
                    dr = np.random.standard_normal(3)
                    dr *= magn/np.sqrt(np.dot(dr,dr))
                    atom = atoms[i]
                    atom.position = atom.position + dr
                    atom = atoms2[i]
                    atom.position = atom.position + dr
                checkenergies(atoms, atoms2)

@pytest.mark.core
@serial
def test_montecarlo_alloy():
    element = "Cu"
    element2 = "Ag"
    latconst = data.reference_states[data.atomic_numbers[element]]['a']
    elementnumber = data.atomic_numbers[element]
    element2number = data.atomic_numbers[element2]
    pbc = tuple(np.random.randint(0,2,3))
    print("Periodic boundaries:", pbc)

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(10,10,10),
                            symbol=element, pbc=pbc)
    atoms2 = atoms.copy()
    atoms = MonteCarloAtoms(atoms)
    print("Number of atoms:", len(atoms))
    # Replace 10% of the atoms with element2
    newz = np.where(np.greater(np.random.random((len(atoms),)), 0.9),
                    element2number, elementnumber)
    atoms.set_atomic_numbers(newz)
    atoms2.set_atomic_numbers(newz)

    atoms.calc = MonteCarloEMT()
    atoms2.calc = EMT()
    atoms.get_potential_energy()

    for e in ((element,elementnumber), (element2,element2number)):
        print ("Number of %s atoms (Z=%d): %d"
            % (e[0], e[1], np.sum(np.equal(atoms.get_atomic_numbers(), e[1]))))

    print() 
    print("Testing perturbations of single atoms")

    magnarr = np.array((0.0, 0.01, 0.1, 1.0, 3.0, 10.0))
    numarr = np.array((1, 3, 10, len(atoms)//2, -3))
    pick1 = np.argsort(np.random.random((len(magnarr),)))
    for magn in np.take(magnarr, pick1):
        pick2 = np.argsort(np.random.random((len(numarr),)))
        for number in np.take(numarr, pick2):
            # Pick number random atoms.
            if number < 0:
                # Find N neighboring atoms and perturb them
                number = -number
                nblist = atoms.calc.get_neighborlist()
                n = 0
                while len(nblist[n]) < number:
                    n += 1
                pick = np.concatenate([np.array((n,)), nblist[n][:number-1]])
                del nblist
            else:
                pick = np.argsort(np.random.random((len(atoms),)))[:number]
            if number > 15:
                s = f"<<< {number} atoms >>>"
            else:
                s = str(list(pick))
            print("  dr = %.3f: %d atom(s): %s" % (magn, number, s))
            for i in pick:
                dr = np.random.standard_normal(3)
                dr *= magn/np.sqrt(np.dot(dr,dr))
                atom = atoms[i]
                atom.position = atom.position + dr
                atom = atoms2[i]
                atom.position = atom.position + dr

            checkenergies(atoms, atoms2)

    print()
    print("Testing alchemy")

    pick2 = np.argsort(np.random.random((len(numarr),)))
    for number in np.take(numarr, pick2):
        # Pick number random atoms.
        if number < 0:
            # Find N neighboring atoms and perturb them
            number = -number
            nblist = atoms.calc.get_neighborlist()
            n = 0
            while len(nblist[n]) < number:
                n += 1
            pick = np.concatenate([np.array((n,)), nblist[n][:number-1]])
            del nblist
        else:
            pick = np.argsort(np.random.random((len(atoms),)))[:number]
        if number > 15:
            s = f"<<< {number} atoms >>>"
        else:
            s = str(list(pick))
            print("  Alchemy, %d atoms: %s" % (number, s))
            for i in pick:
                atom = atoms[i]
                atom2 = atoms2[i]
                z = atom.number
                assert z == atom2.number
                if (z == elementnumber):
                    atom.number = element2number
                    atom2.number = element2number
                else:
                    assert z == element2number
                    atom.number = elementnumber
                    atom2.number = elementnumber
            checkenergies(atoms, atoms2)


def checklist(nb, a, cut):
    error = 0
    a2 = Atoms(a)
    nb2 = FullNeighborList(cut, a2, 0)
    print("    Checking Cartesian positions")
    err = abs((a.get_positions() - a2.get_positions()).ravel())
    idx = np.argmax(err)
    at, co = idx//3, idx%3
    print("      err =", err[idx], "at", idx, ":", at, co)
    ReportTest("      Worst Cartesian position (%d,%d)" % (at, co),
               a.get_positions()[at,co],
               a2.get_positions()[at,co], 1e-10)

    print("    Checking full list")
    for i in range(len(a)):
        l1 = list(nb[i])
        l1.sort()
        l2 = list(nb2[i])
        l2.sort()
        if not l1 == l2:
            ReportTest.BoolTest("NB lists for atom %d should be identical"
                                % (i,), False)
            print("nb1[%d] = %s" % (i, str(l1)))
            print("nb2[%d] = %s" % (i, str(l2)))
            print("pos1[%d] = %s" % (i, str(a.get_positions()[i])))
            print("pos2[%d] = %s" % (i, str(a2.get_positions()[i])))
            error += 1
            if error > 3:
                print("Too many errors, giving up!")
                raise RuntimeError("Too many errors, giving up!")


@serial
def test_montecarlo_nblist():
    listcutoff = 4.98409   # The EMT value for Cu
    element = "Cu"
    latconst = data.reference_states[data.atomic_numbers[element]]['a']
    pbc = tuple(np.random.randint(0,2,3))
    print("Periodic boundaries:", pbc)

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(10,10,10),
                            symbol=element, pbc=pbc, debug=0)

    nblist = NeighborList(listcutoff, atoms, 0, full=True)
    print()
    print("Checking number of affected atoms.")
    #Verbose(1)

    np_atoms = Atoms(atoms, pbc=False)
    np_nblist = NeighborList(listcutoff, np_atoms, 0, full=True)

    i = np.random.randint(0, len(np_atoms))
    dr = np.random.standard_normal(3)
    dr *= 10.0/np.sqrt(np.dot(dr,dr))
    atom = np_atoms[i]
    oldpos = np.array(atom.position)
    newpos = oldpos + dr
    atom.position = newpos
    nr = np_nblist.test_partial_update(np.array([i,], np.int32), np_atoms)
    # Count who should be affected.
    r = np_atoms.get_positions() - oldpos
    r2 = np.sum(r * r, 1)
    a = np.less_equal(r2, listcutoff*listcutoff)
    r = np_atoms.get_positions() - newpos
    r2 = np.sum(r * r, 1)
    b = np.less_equal(r2, listcutoff*listcutoff)

    nr2 = np.sum(np.logical_or(a,b))
    print(nr, "atoms were affected")
    print(np.sum(a), "near old position and", np.sum(b), "near new position.")
    print(nr2, "atoms should be affected")
    ReportTest("Number of affected atoms", nr, nr2, 0)
    
    del np_atoms, np_nblist
    
    print()
    print("Testing perturbations of all the atoms")

    magnarr = np.array((0.0, 0.01, 0.1, 1.0, 3.0, 10.0))
    pick = np.argsort(np.random.random((len(magnarr),)))
    for magn in np.take(magnarr, pick):
        dx = magn * np.random.uniform(-1, 1, (len(atoms), 3))
        print("  Perturbation:", magn)
        atoms.set_positions(dx + atoms.get_positions())
        nblist.check_and_update(atoms)
        checklist(nblist, atoms, listcutoff)

    print() 
    print("Testing perturbations of single atoms")

    magnarr = np.array((0.0, 0.01, 0.1, 1.0, 3.0, 10.0))
    pick1 = np.argsort(np.random.random((len(magnarr),)))
    for magn in np.take(magnarr, pick1):
        numarr = np.array((1, 3, 10, len(atoms)//2, -3))
        pick2 = np.argsort(np.random.random((len(numarr),)))
        for number in np.take(numarr, pick2):
            # Pick number random atoms.
            if number < 0:
                # Find N neighboring atoms and perturb them
                number = -number
                n = 0
                while len(nblist[n]) < number:
                    n += 1
                pick = np.concatenate([np.array((n,)), nblist[n][:number-1]])
            else:
                pick = np.argsort(np.random.random((len(atoms),)))[:number]
            if number > 15:
                s = f"<<< {number} atoms >>>"
            else:
                s = str(list(pick))
            print("  dr = %.3f: %d atoms: %s" % (magn, number, s))
            for i in pick:
                atom = atoms[i]
                dr = np.random.standard_normal(3)
                dr *= magn/np.sqrt(np.dot(dr,dr))
                atom.position += dr

            nblist.test_partial_update(pick.astype(np.int32), atoms)
            checklist(nblist, atoms, listcutoff)



import numpy as np
from asap3.test.nvt_module import main
from asap3.md.nvtberendsen import NVTBerendsen

from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

## @pytest.mark.slow    # Not marked as slow so a single nvt test runs normally
@serial
def test_nvt_berendsen():
    def berdynmaker(atoms, T0, dt, tau, rng, logint):
        return NVTBerendsen(atoms, timestep=dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()
    main(berdynmaker, rng, failfluct=True)

@pytest.mark.slow
@parallel
def test_par_nvt_berendsen():
    def berdynmaker(atoms, T0, dt, tau, rng, logint):
        return NVTBerendsen(atoms, timestep=dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint)

    #rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    rng = np.random.default_rng()
    main(berdynmaker, rng, failfluct=True,  parallel=True)



"""Asap builtin objects.

This module loads the modules from the compiled parts of Asap.  It
chooses between the parallel and serial version based on whether MPI
is active at runtime.
"""

__docformat__ = "restructuredtext en"

import sys

# Now import the relevant C module
import _asap
        
parallelpossible = hasattr(_asap, "Communicator")
AsapError = _asap.AsapError
get_version = _asap.get_version
get_short_version = _asap.get_short_version
timing_results = _asap.timing_results

def set_verbose(v):
    """Set the verbosity level of asap."""
    _asap.verbose = v
    



"""Asap module ParallelListOfAtoms.

Defines the parallel list of atoms object (`ParallelAtoms`), and a factory
method for creating them (`MakeParallelAtoms`).

Importing this module also installs a Python exit function causing
MPI_Abort to be called if an uncaught exception occurs.
"""


__docformat__ = "restructuredtext en"

import ase
import ase.units
try:
    from ase.geometry.cell import Cell
except ImportError:
    # Cell was introduced in ASE in April 2019.
    Cell = None
from asap3 import _asap
from asap3.Internal.BuiltinPotentials import AsapPotential
import asap3.mpi
import numpy as np
import pickle
import sys, time
import ase.parallel
import numbers
from ase.utils import deprecated
import copy


class BorrowedNeighborLocator:
    """A front for a neighbor locator borrowed from a parallel potential.

    In a parallel simulation, borrowing a neighbor locator from a 
    potential and then updating it will corrupt the potential as
    a migration may be triggered without the potentials knowledge.

    This front for the borrowed neighbor list solves that problem
    by asking the potential to do the update.
    """
    def __init__(self, nblist, potential):
        self._nblist = nblist
        self._potential = potential

    def check_and_update(self, atoms):
        self._potential.check_update_neighborlist(atoms)

    # All other methods are passed onto the neighbor list
    def __getattr__(self, name):
        return getattr(self._nblist, name)


class ParallelPotential(_asap.ParallelPotential, AsapPotential):
    def __init__(self, potential, *args, **kwargs):
        super().__init__(potential, *args, **kwargs)
        self._calc = potential
        
    def get_stress(self, atoms):
        stress = self.get_virial(atoms)
        if not getattr(atoms, "_ase_handles_dynamic_stress", False):
            p = atoms.get_momenta()
            masses = atoms.get_masses()
            invmass = 1.0 / masses
            dynstress = np.zeros_like(stress)
            for alpha in range(3):
                for beta in range(alpha, 3):
                    dynstress[self._stresscomp[alpha,beta]] = -(p[:,alpha] * p[:,beta] * invmass).sum()
            asap3.mpi.world.sum(dynstress)
            stress += dynstress
        stress /= atoms.get_volume()
        return stress

    def get_neighborlist(self):
        nbl = self._calc.get_neighborlist()
        return BorrowedNeighborLocator(nblist=nbl, potential=self)

    # Some Potentials may have non-standard methods and attributes.
    # In those cases, pass through to the original Potential.
    def __getattr__(self, name):
        return getattr(self._calc, name)

class NoCalculatorWrapper:
    """Object proxy hiding the calculator of an Atoms object."""
    def __init__(self, wrappee):
        self._wrappee = wrappee

    def get_calculator(self):
        return None

    def __getattr__(self, attr):
        return getattr(self._wrappee, attr)

    @property
    def calc(self):
        """Calculator object."""
        return None



class ParallelAtoms(ase.Atoms):
    """Atoms class for parallel Asap simulations.

    It is recommended to create ParallelAtoms objects using
    `MakeParallelAtoms`.
    """
    parallel = 1
    def __init__(self, nCells, comm, atoms, cell=None, pbc=None,
                 distribute=True):
        """Create a ParallelAtoms object.

        WARNING: ParallelAtoms object should normally not be created
        explicitly.  Use MakeParallelAtoms instead.
        """
        # Initialize the atoms.  Hide the calculator: Often there will
        # be a SinglePointCalculator (or anothter calculator) on the
        # master but nothing on the slaves, causing a deadlock.  And a
        # ParallelPotential cannot be reused on a new ParallelAtoms
        # object.
        super().__init__(NoCalculatorWrapper(atoms), cell=cell, pbc=pbc, constraint=None)

        # Sanity checks
        assert self.arrays["positions"].dtype == np.dtype(float)
        assert self.arrays["positions"].shape == (len(atoms), 3)
        assert self.arrays["numbers"].shape ==  (len(atoms),)

        # The initializer of the parent class (ase.Atoms) only copies
        # known arrays.  Copy the rest.
        for k, v in atoms.arrays.items():
            if k not in self.arrays:
                self.arrays[k] = v.copy()

        self.nCells = np.array(nCells, int)
        if self.nCells.shape != (3,):
            raise ValueError("ParallelAtoms: nCells must be 3 integers.")
        self.comm = comm

        self.ghosts = {}
        self.ghosts["positions"] = np.zeros((0,3), float)
        self.ghosts["numbers"] = np.zeros(0, self.arrays["numbers"].dtype)

        # Now make the IDs
        mynatoms = np.array([len(self)])
        natoms_all = np.zeros(self.comm.size, int)
        self.comm.all_gather(mynatoms, natoms_all)
        if "ID" not in self.arrays:
            firstID = sum(natoms_all[:self.comm.rank])
            self.arrays["ID"] = np.arange(firstID, firstID+len(atoms))
        self.total_number_of_atoms = sum(natoms_all)

        if distribute:
            self.distribute()

    def distribute(self, verbose=None):
        if verbose is None:
            verbose = _asap.verbose
        _asap.DistributeAtoms(self, verbose)

    def get_global_number_of_atoms(self):
        n = len(self)
        return self.comm.sum(n)

    def get_number_of_degrees_of_freedom(self):
        n = super().get_number_of_degrees_of_freedom()
        return self.comm.sum(n)
    
    def get_list_of_elements(self):
        """Get a list of elements.

        The list is cached to prevent unnecessary communication.
        """
        try:
            return self.listofelements
        except AttributeError:
            z = self.get_atomic_numbers()
            present = np.zeros(100, int)
            if len(z):
                zmax = z.max()
                zmin = z.min()
                present[zmin] = present[zmax] = 1
                for i in range(zmin+1, zmax):
                    if np.equal(z, i).any():
                        present[i] = 1
            self.comm.sum(present)
            self.listofelements = []
            for i, p in enumerate(present):
                if p:
                    self.listofelements.append(i)
            return self.listofelements

    def set_atomic_numbers(self, numbers):
        """Set the atomic numbers."""
        try:
            # Discard the cached list of elements
            del self.listofelements
        except AttributeError:
            pass
        ase.Atoms.set_atomic_numbers(self, numbers)

    def get_ids(self):
        """Get the atom IDs in a parallel simulation."""
        return self.arrays["ID"].copy()

    def is_master(self):
        """Return 1 on the master node, 0 on all other nodes."""
        return (self.comm.rank == 0)

    def get_comm(self):
        return self.comm
    
    def wrap_calculator(self, calc):
        "Make an ASAP calculator compatible with parallel simulations."
        try:
            parallelOK = calc.supports_parallel()
        except AttributeError:
            parallelOK = False
        if not parallelOK:
            raise ValueError("The calculator does not support parallel ASAP calculations.")
        try:
            verbose = calc.verbose
        except AttributeError:
            verbose = _asap.verbose
        return ParallelPotential(calc, verbose)

    def set_calculator(self, calc, wrap=True):
        """Sets the calculator in a way compatible with parallel simulations.
        
        calc: 
            The Calculator to be used.  Normally only Asap calculators will work.
            
        wrap (optional, default=True):
            Indicates if a calculator should be wrapped in a ParallelCalculator object.  
            Wrapping is the default, and should almost always be used, the only exception
            being if the Calculator is implemented as a Python object wrapping an Asap
            calculator, in which case the Asap calculator should first be wrapped in
            a ParallelCalculator object (use atoms.wrap_calculator) and this one should then
            be used by the Python calculator.  The Python calculator is then attached
            without being wrapped again.
        """
        if wrap and calc is not None:
            parcalc = self.wrap_calculator(calc)
        else:
            parcalc = calc
        ase.Atoms.calc.fset(self, parcalc)

    def copy(self):
        """Return a copy."""
        atoms = self.__class__(self.nCells, self.comm, self)
        atoms.constraints = copy.deepcopy(self.constraints)
        return atoms

    @property
    def calc(self):
        """Calculator object."""
        return ase.Atoms.calc.fget(self)
    
    @calc.setter
    def calc(self, calc):
        self.set_calculator(calc)

    @calc.deleter  # type: ignore
    @deprecated(DeprecationWarning('Please use atoms.calc = None'))
    def calc(self):
        self.set_calculator(None)

    def get_kinetic_energy(self):
        local_ekin = ase.Atoms.get_kinetic_energy(self)
        return self.comm.sum(local_ekin)
    
    def get_temperature(self):
        """Get the temperature. in Kelvin"""
        dof = 3 * self.get_global_number_of_atoms()
        removed_dof = 0
        for c in self.constraints:
            removed_dof += c.get_removed_dof(self)
        removed_dof = self.comm.sum(int(removed_dof))
        ekin_per_dof = self.get_kinetic_energy() / (dof - removed_dof)
        return ekin_per_dof / (0.5 * ase.units.kB)

    def get_ghost_positions(self):
        return self.ghosts['positions'].copy()
    
    def get_ghost_atomic_numbers(self):
        return self.ghosts['numbers'].copy()
    
    def get_center_of_mass(self, scaled=False):
        """Get the center of mass.

        If scaled=True the center of mass in scaled coordinates
        is returned."""
        m = self.get_masses()
        rsum = np.dot(m, self.arrays['positions'])
        msum = m.sum()
        data = np.zeros(4)
        data[:3] = rsum
        data[3] = msum
        self.comm.sum(data)
        com = data[:3] / data[3]
        if scaled:
            return np.linalg.solve(self._cell.T, com)
        else:
            return com


    def get_moments_of_inertia(self, vectors=False):
        """Get the moments of inertia along the principal axes.

        The three principal moments of inertia are computed from the
        eigenvalues of the symmetric inertial tensor. Periodic boundary
        conditions are ignored. Units of the moments of inertia are
        amu*angstrom**2.
        """
        com = self.get_center_of_mass()
        positions = self.get_positions()
        positions -= com  # translate center of mass to origin
        masses = self.get_masses()

        # Initialize elements of the inertial tensor
        I11 = I22 = I33 = I12 = I13 = I23 = 0.0
        for i in range(len(self)):
            x, y, z = positions[i]
            m = masses[i]

            I11 += m * (y ** 2 + z ** 2)
            I22 += m * (x ** 2 + z ** 2)
            I33 += m * (x ** 2 + y ** 2)
            I12 += -m * x * y
            I13 += -m * x * z
            I23 += -m * y * z

        I = np.array([[I11, I12, I13],
                      [I12, I22, I23],
                      [I13, I23, I33]])
        self.comm.sum(I)

        evals, evecs = np.linalg.eigh(I)
        if vectors:
            return evals, evecs.transpose()
        else:
            return evals

    def get_angular_momentum(self):
        """Get total angular momentum with respect to the center of mass."""
        data = super().get_angular_momentum()
        self.comm.sum(data)
        return data
    
    def get_stress(self, voigt=True, apply_constraint=True, include_ideal_gas=False):
        """Calculate stress tensor.

        Returns an array of the six independent components of the
        symmetric stress tensor, in the traditional Voigt order
        (xx, yy, zz, yz, xz, xy) or as a 3x3 matrix.  Default is Voigt
        order.

        The ideal gas contribution to the stresses is added if the 
        atoms have momenta, unless it is explicitly disabled.
        """

        if self._calc is None:
            raise RuntimeError('Atoms object has no calculator.')

        stress = self._calc.get_stress(self)
        assert stress.shape == (6,)

        if apply_constraint:
            for constraint in self.constraints:
                if hasattr(constraint, 'adjust_stress'):
                    constraint.adjust_stress(self, stress)

        # Add ideal gas contribution, if applicable
        if getattr(self, "_ase_handles_dynamic_stress", False) and include_ideal_gas and self.has('momenta'):
            stresscomp = np.array([[0, 5, 4], [5, 1, 3], [4, 3, 2]])
            invvol = 1.0 / self.get_volume()
            p = self.get_momenta()
            masses = self.get_masses()
            invmass = 1.0 / masses
            dynstress = np.zeros_like(stress)
            for alpha in range(3):
                for beta in range(alpha, 3):
                    dynstress[stresscomp[alpha,beta]] -= (p[:,alpha] * p[:,beta] * invmass).sum() * invvol
            asap3.mpi.world.sum(dynstress)
            stress += dynstress

        if voigt:
            return stress
        else:
            xx, yy, zz, yz, xz, xy = stress
            return np.array([(xx, xy, xz),
                             (xy, yy, yz),
                             (xz, yz, zz)])

    def delete_atoms_globally(self, local_indices=None, global_indices=None):
        """Delete atoms from a parallel simulation.

        local_indices: Remove atoms with these indices on the local MPI task.

        global_indices: Remove atoms with these global IDs.  The atoms
        may reside on this or on another MPI task.

        Both parameters should be sequences, or a single int.  Both can
        be specified simultaneously, if need be, although that may be
        confusing.

        This method must be called simultaneously by all MPI tasks, but
        the lists of indices may be diffent on the different tasks.  An
        atom is deleted if it is specified in at least one task.
        Duplicates are allowed.
        """
        for c in self.constraints:
            # Constraint must have data stored on atoms.
            c.start_asap_dynamics(self)
        indices = []
        if isinstance(local_indices, int):
            local_indices = [local_indices]
        if isinstance(global_indices, int):
            global_indices = [global_indices]
        if local_indices is not None and len(local_indices):
            # Convert to global indices
            ids = self.get_ids()
            indices += [ids[i] for i in local_indices]
        if global_indices is not None and len(global_indices):
            indices += list(global_indices)
        # Now merge all these lists.  They may be empty on most MPI tasks.
        number = len(indices)
        number = self.comm.max(number)
        if number == 0:
            # Not deleting anything is valid.
            return
        natoms = self.get_global_number_of_atoms()
        indices_array = -np.ones(number, int)  # Array of the value -1
        indices_array[:len(indices)] = indices
        merged_indices = np.zeros(number * self.comm.size, int)
        self.comm.all_gather(indices_array, merged_indices)
        # Now all tasks have a list of the global indices of all atoms
        # to delete, polluted by elements with value -1 and by
        # duplicates.  Clean it up.
        merged_indices = np.unique(merged_indices[merged_indices != -1])

        # Ready to delete.  We need to "prune" the arrays on some tasks,
        # and to make sure that the IDs remain contiguous.  We start by
        # swapping IDs between the atoms being removed, and the ones
        # with the largest IDs.  Then we remove the ones with the
        # largest IDs.  It is possible that some atoms to be removed
        # have so large IDs that they do not need an ID swap.
        remove_above = natoms - len(merged_indices)
        to_be_swapped = list(range(remove_above, natoms))
        print(merged_indices, remove_above, to_be_swapped)
        for idx in merged_indices:
            if idx in to_be_swapped:
                to_be_swapped.remove(idx)
        swp = {}
        for idx in merged_indices:
            if idx < remove_above:
                swp[idx] = to_be_swapped.pop()
        assert len(to_be_swapped) == 0
        ids = self.arrays["ID"]
        for k, v in swp.items():
            # Now swap k and v in arrays["ID"].   They may or may not be present.
            k_idx = np.nonzero(ids == k)
            v_idx = np.nonzero(ids == v)
            if len(k_idx) == 1:
                ids[k_idx[0]] = v
            if len(v_idx) == 1:
                ids[v_idx[0]] = k
        # Now the atoms to be removed have ids of remove_above or higher
        keep = ids < remove_above
        if keep.sum() < len(keep):
            for k, v in self.arrays.items():
                self.arrays[k] = v[keep]
        assert self.get_global_number_of_atoms() == natoms - len(merged_indices)

        # Ghost atoms are now incorrect.  Better delete them, so accessing them cause a visible error
        self._asap_invalidghosts = True
        for k, v in self.ghosts.items():
            self.ghosts[k] = np.zeros((0,) + v.shape[1:], v.dtype)

        # Final sanity check: Any constraint must have its data on the atoms, otherwise
        # it will now be corrupt.
        for constraint in self.constraints:
            if not getattr(constraint, 'asap_ready', False):
                raise ValueError("Deleting atoms has corrupted contraint "+str(constraint))
        for c in self.constraints:
            c.end_asap_dynamics(self)

    # It is not possible to delete atoms with __delitem__ as that cannot sync correctly.
    def __delitem__(self, *args, **kwargs):
        raise NotImplementedError("Cannot delete parallel atoms with del.  Use atoms.delete_atoms_globally()")
        
    # We need to redefine __getitem__ since __init__ does not
    # take the expected arguments.
    def __getitem__(self, i):
        """Return a subset of the atoms.

        i -- scalar integer, list of integers, or slice object
        describing which atoms to return.

        If i is a scalar, return an Atom object. If i is a list or a
        slice, return an Atoms object with the same cell, pbc, and
        other associated info as the original Atoms object. The
        indices of the constraints will be shuffled so that they match
        the indexing in the subset returned.

        The returned object is an ordinary Atoms object, not a
        ParallelAtoms object.
        """
        if isinstance(i, numbers.Integral):
            natoms = len(self)
            if i < -natoms or i >= natoms:
                raise IndexError('Index out of range.')

            return ase.Atom(atoms=self, index=i)

        import copy
        from ase.constraints import FixConstraint, FixBondLengths

        atoms = ase.Atoms(cell=self._cell, pbc=self._pbc, info=self.info)

        atoms.arrays = {}
        for name, a in self.arrays.items():
            atoms.arrays[name] = a[i].copy()

        # Constraints need to be deepcopied, since we need to shuffle
        # the indices
        atoms.constraints = copy.deepcopy(self.constraints)
        condel = []
        for con in atoms.constraints:
            if isinstance(con, (FixConstraint, FixBondLengths)):
                try:
                    con.index_shuffle(self, i)
                except IndexError:
                    condel.append(con)
        for con in condel:
            atoms.constraints.remove(con)
        return atoms

def MakeParallelAtoms(atoms, nCells, cell=None, pbc=None,
                      distribute=True):
    """Build parallel simulation from serial lists of atoms.

    Call simultaneously on all processors.  Each processor having
    atoms should pass a list of atoms as the first argument, or None
    if this processor does not contribute with any atoms.  If the
    cell and/or pbc arguments are given, they must be given on
    all processors, and be identical.  If it is not given, a supercell
    is attempted to be extracted from the atoms on the processor with
    lowest rank.

    Any atoms object passed to this method cannot have a calculator or a
    contraint attached.

    This is the preferred method for creating parallel simulations.
    """
    mpi = asap3.mpi
    #comm = mpi.world.duplicate()
    comm = mpi.world

    # Sanity check: No contraint.
    if atoms is not None:
        if getattr(atoms, "contraints", None):
            raise ValueError("The atoms on node {} have contraints: {}".format(
                comm.rank, str(atoms.constraints)))

    # Sanity check: is the node layout reasonable
    nNodes = nCells[0] * nCells[1] * nCells[2]
    if nNodes != comm.size:
        raise RuntimeError("Wrong number of CPUs: %d != %d*%d*%d" %
                           (comm.size, nCells[0], nCells[1], nCells[2]))
    t1 = np.zeros((3,))
    t2 = np.zeros((3,))
    comm.min(t1)
    comm.max(t2)
    if (t1[0] != t2[0] or t1[1] != t2[1] or t1[2] != t2[2]):
        raise RuntimeError("CPU layout inconsistent.")

    # If pbc and/or cell are given, they may be shorthands in need of
    # expansion.
    if pbc:
        try:
            plen = len(pbc)
        except TypeError:
            # It is a scalar, interpret as a boolean.
            if pbc:
                pbc = (1,1,1)
            else:
                pbc = (0,0,0)
        else:
            if plen != 3:
                raise ValueError("pbc must be a scalar or a 3-sequence.")
    if cell:
        cell = array(cell)  # Make sure it is a numeric array.
        if cell.shape == (3,):
            cell = array([[cell[0], 0, 0],
                          [0, cell[1], 0],
                          [0, 0, cell[2]]])
        elif cell.shape != (3,3):
            raise ValueError("Unit cell must be a 3x3 matrix or a 3-vector.")

    # Find the lowest CPU with atoms, and let that one distribute
    # which data it has.  All other CPUs check for consistency.
    if atoms is None:
        hasdata = None
        mynum = comm.size
    else:
        hasdata = {}
        for name in atoms.arrays.keys():
            datatype = atoms.arrays[name].dtype.char
            shape = atoms.arrays[name].shape[1:]
            hasdata[name] = (datatype, shape)
        mynum = comm.rank
        if pbc is None:
            pbc = atoms.get_pbc()
        if cell is None:
            cell = np.array(atoms.get_cell())
    root = comm.min(mynum)   # The first CPU with atoms
    # Now send hasdata, cell and pbc to all other CPUs
    package = pickle.dumps((hasdata, cell, pbc), 2)
    package = comm.broadcast_string(package, root)
    rootdata, rootcell, rootpbc = pickle.loads(package)
    if rootdata is None or len(rootdata) == 0:
        raise ValueError("No data from 'root' atoms.  Empty atoms?!?")
    
    # Check for consistent cell and pbc arguments
    if cell is not None:
        if rootcell is None:
            raise TypeError("Cell given on another processor than the atoms.")
        if (cell.ravel() - rootcell.ravel()).max() > 1e-12:
            raise ValueError("Inconsistent cell specification.")
    else:
        cell = rootcell   # May still be None
    if pbc is not None:
        if rootpbc is None:
            raise TypeError("PBC given on another processor than the atoms.")
        if (pbc != rootpbc).any():
            raise ValueError("Inconsistent pbc specification.")
    else:
        pbc = rootpbc

    # Check for consistent atoms data
    if hasdata is not None:
        if hasdata != rootdata:
            raise ValueError("Atoms do not contain the sama data on different processors.")
    if "positions" not in rootdata:
        raise ValueError("Atoms do not have positions!")
    
    # Create empty atoms
    if atoms is None:
        atoms = ase.Atoms(cell=cell, pbc=pbc)
        for name in rootdata.keys():
            if name in atoms.arrays:
                assert atoms.arrays[name].dtype.char == rootdata[name][0]
                assert len(atoms.arrays[name]) == 0
            else:
                shape = (0,) + rootdata[name][1]
                atoms.arrays[name] = np.zeros(shape, rootdata[name][0])
        
    return ParallelAtoms(nCells, comm, atoms, cell=cell, pbc=pbc, 
                         distribute=distribute)



# A cleanup function should call MPI_Abort if python crashes to
# terminate the processes on the other nodes.
ase.parallel.register_parallel_cleanup_function()

# _oldexitfunc = getattr(sys, "exitfunc", None)
# def _asap_cleanup(lastexit = _oldexitfunc, sys=sys, time=time,
#                   comm = asap3.mpi.world):
#     error = getattr(sys, "last_type", None)
#     if error:
#         sys.stdout.flush()
#         sys.stderr.write("ASAP CLEANUP (node " + str(comm.rank) +
#                          "): " + str(error) +
#                          " occurred.  Calling MPI_Abort!\n")
#         sys.stderr.flush()
#         # Give other nodes a moment to crash by themselves (perhaps
#         # producing helpful error messages).
#         time.sleep(3)
#         comm.abort(42)
#     if lastexit:
#         lastexit()
# sys.exitfunc = _asap_cleanup
        
# END OF PARALLEL STUFF



'''Test that the Asap mpi interface matches GPAw's - not currently the case.

The Asap MPI module is missing some functions, that causes this to raise
exceptions, which causes MPI to abort.  It is not enough to mark the
functions as xfail, they need to be skipped completely.
'''
import pytest
import numpy as np
from asap3.mpi import world # broadcast_array # , send, receive
from asap3.test.pytest_markers import parallel

import pytest

@parallel
@pytest.mark.skip
def test_send_receive_object():
    if world.size == 1:
        return
    obj = (42, 'hello')
    if world.rank == 0:
        send(obj, 1, world)
    elif world.rank == 1:
        assert obj == receive(0, world)


@parallel
def test_scalar_reduce():
    'Gpaw-style scalar reduction functions.'
    assert world.sum_scalar(world.rank + 1) == world.size * \
        (world.size + 1) // 2
    assert np.allclose(world.sum_scalar(world.rank + 1.0),
                       world.size * (world.size + 1.0) / 2)
    assert world.min_scalar(world.rank + 1) == 1
    assert world.max_scalar(world.rank + 1) == world.size
    assert world.min_scalar(world.rank + 1.0) == 1.0
    assert world.max_scalar(world.rank + 1.0) == world.size * 1.0

@parallel
def test_implicit_scalar_reduce():
    'Asap mpi object can reduce scalars with the same function as arrays'
    assert world.sum(world.rank + 1) == world.size * \
        (world.size + 1) // 2
    assert np.allclose(world.sum(world.rank + 1.0),
                       world.size * (world.size + 1.0) / 2)
    assert world.min(world.rank + 1) == 1
    assert world.max(world.rank + 1) == world.size
    assert world.min(world.rank + 1.0) == 1.0
    assert world.max(world.rank + 1.0) == world.size * 1.0


@parallel
@pytest.mark.skip
def test_bcast_array():
    'Broadcast of arrays on multiple communicators simultanously - not needed in Asap'
    new = world.new_communicator

    if world.size == 2:
        comms = [world]
    elif world.size == 4:
        ranks = np.array([[0, 1], [2, 3]])
        comms = [new(ranks[world.rank // 2]),
                 new(ranks[:, world.rank % 2])]
    elif world.size == 8:
        ranks = [[[0, 1], [0, 2], [0, 4]],
                 [[0, 1], [1, 3], [1, 5]],
                 [[2, 3], [0, 2], [2, 6]],
                 [[2, 3], [1, 3], [3, 7]],
                 [[4, 5], [4, 6], [0, 4]],
                 [[4, 5], [5, 7], [1, 5]],
                 [[6, 7], [4, 6], [2, 6]],
                 [[6, 7], [5, 7], [3, 7]]]
        comms = [new(r) for r in ranks[world.rank]]
    else:
        return

    array = np.zeros(3, int)
    if world.rank == 0:
        array[:] = 42

    out = broadcast_array(array, *comms)
    assert (out == 42).all()



from asap3 import EMT, EMT2013, Morse, LennardJones
from asap3.EMT2013Parameters import sihb_PtY_parameters as PtY_EMT2013
import asap3
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import L1_2
from ase.data import atomic_numbers, reference_states
from asap3.test.pytest_markers import parallel
from ase.parallel import world

import numpy as np

import os
import pickle

import pytest

pytestmark = [pytest.mark.core, ]

@pytest.fixture
def results(datadir):
    resultname = 'parallelPotentials.pickle'
    with open(datadir / resultname, 'rb') as resfile:
        return pickle.load(resfile, encoding='latin1')
    
def ljpot():
    return LennardJones([29], [0.15], [2.7], 3*2.7, False)

def morse():
    return Morse(
        elements = np.array([atomic_numbers['Ru'], atomic_numbers['Ar']]),
        epsilon = np.array([[5.720, 0.092], [0.092, 0.008]]),
        alpha = np.array([[1.475, 2.719], [2.719, 1.472]]),
        rmin = np.array([[2.110, 2.563], [2.563, 4.185]])
    )

parameters = [
    # name, element(s), pbc, calculator, stress
    ['Cu EMT periodic', 'Cu', True, EMT(), True],
    ['Cu EMT free', 'Cu', False, EMT(), True],
    ['Cu EMT mixed', 'Cu', (1,1,0), EMT(), True],
    ['AuCu3 EMT periodic', ('Au', 'Cu'), True, EMT(), True],
    ['AuCu3 EMT free', ('Au', 'Cu'), False, EMT(), True],
    ['LennardJones periodic', 'Cu', True, ljpot(), True],
    ['LennardJones free', 'Cu', False, ljpot(), True],
    ['LennardJones mixed', 'Cu', (1,1,0), ljpot(), True],
    ['Morse periodic', 'Ar', True, morse(), False],
    ['Morse free', 'Ar', False, morse(), False],
    ['Morse mixed', 'Ar', (0,1,0), morse(), False],
    ['EMT2013 Pt3Y periodic', ('Y', 'Pt'), True, EMT2013(PtY_EMT2013), True],
    ['EMT2013 Pt3Y free', ('Y', 'Pt'), False, EMT2013(PtY_EMT2013), True],
    ['EMT2013 Pt3Y mixed', ('Y', 'Pt'), (0,0,1), EMT2013(PtY_EMT2013), True],
]

@pytest.mark.skipif(world.size not in (1, 2, 4, 8), reason='Number of cores must be 1,2,4 or 8')
@pytest.mark.parametrize('name,elements,pbc,calculator,stress',
                         parameters, ids=[x[0] for x in parameters])
def test_potential_par(name, elements, pbc, calculator, stress, results, cpulayout):
    if isinstance(elements, tuple):
        atoms = myalloy(elements, pbc, cpulayout)
    else:
        atoms = myfcc(elements, pbc, cpulayout)
    atoms.calc = calculator
    dotest(atoms, name, results, dostress=stress)

def myfcc(symbol, pbc, cpulayout):
    "Make an fcc lattice with standard lattice constant"
    if world.rank == 0:
        a = FaceCenteredCubic(symbol=symbol, size=(10,10,10), pbc=pbc)    
        dx = 0.01 * np.sin(0.1 * np.arange(len(a) * 3))
        dx.shape = (len(a),3)
        a.set_positions(a.get_positions() + dx)
        a.set_momenta(np.zeros((len(a),3)))
    else:
        a = None
    if world.size > 1:
        a = asap3.MakeParallelAtoms(a, cpulayout)
    return a

def myalloy(symbols, pbc, cpulayout):
    "Create an L1_2 alloy (AB3)."
    a1 = reference_states[atomic_numbers[symbols[0]]]['a']
    a2 = reference_states[atomic_numbers[symbols[1]]]['a']
    a0 = (a1 + 3*a2)/4
    if world.rank == 0:
        a = L1_2(symbol=symbols, size=(10,10,10), latticeconstant=a0, pbc=pbc)    
        dx = 0.01 * np.sin(0.1 * np.arange(len(a) * 3))
        dx.shape = (len(a),3)
        a.set_positions(a.get_positions() + dx)
        a.set_momenta(np.zeros((len(a),3)))
    else:
        a = None
    if world.size > 1:
        a = asap3.MakeParallelAtoms(a, cpulayout)
    return a
    
    
def dotest(atoms, name, results, dostress=True):
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    positions = atoms.get_positions()
    if dostress:
        stresses = atoms.get_stresses()
        stress = atoms.get_stress()
    if world.size > 1:
        ids = atoms.get_ids()
    else:
        ids = np.arange(len(atoms))
    
    # Test that we still get the same
    data = results[name]
    assert np.isclose(energy, data['energy'], atol=1e-9), name + ' (energy)'
    compare(name, 'forces', forces, data['forces'], ids)
    compare(name, 'positions', positions, data['positions'], ids)
    if dostress:
        for i in range(6):
            assert np.isclose(stress[i], data['stress'][i], atol=1e-9), name + f'(stress {i})' 
        compare(name, 'stresses', stresses, data['stresses'], ids)

        
def compare(name, quantity, a, b, ids):
    keep = 50   # Keep every 50 data point
    for i in range(len(a)):
        ii = ids[i]
        if ii % keep == 0:
            x = a[i]
            y = b[ii // keep]
            for j in range(len(x)):
                testname = '%i %s (%s[%i,%i])' % (world.rank, name, quantity, ii, j)
                assert np.isclose(x[j], y[j], atol=1e-10), testname




from asap3 import EMT, Atoms
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest
from asap3.test.pytest_markers import serial

import pytest

@serial
@pytest.mark.core
def test_changecalculator():
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                            size=(6,6,6), symbol="Cu")
    atoms.calc = EMT()
    f1 = atoms.get_forces()
    atoms.calc = EMT()
    f2 = atoms.get_forces()
    maxdev = abs(f2 - f1).max()
    print(maxdev)
    ReportTest("Max error 1:", maxdev, 0.0, 1e-6)

    atoms2 = Atoms(atoms)

    f2 = atoms2.get_forces()
    maxdev = abs(f2 - f1).max()
    print(maxdev)
    ReportTest("Max error 2:", maxdev, 0.0, 1e-6)

    f2 = atoms.get_forces()
    maxdev = abs(f2 - f1).max()
    print(maxdev)
    ReportTest("Max error 1:", maxdev, 0.0, 1e-6)




from asap3 import *
from asap3.mpi import world
from ase.lattice.cubic import *
from asap3.testtools import ReportTest
import time
import os

debug = 0
if debug == 1:
    DebugOutput("makeparallel%d.log", nomaster=True)
elif debug == 2:
    time.sleep(world.rank)
    print("PID:", os.getpid())
    time.sleep(20)

if world.size == 1:
    layout = None
elif world.size == 2:
    layout = [2,1,1]
elif world.size == 3:
    layout = [1,3,1]
elif world.size == 4:
    layout = [2,1,2]

def report():
    for i in range(world.size):
        if i == world.rank:
            print("Data on processor", i)
            for key in atoms.arrays.keys():
                print("  ", key, atoms.arrays[key].shape)
            r = atoms.get_positions()
            if len(r):
                print("Limits to the positions:")
                print ("[%.4f, %.4f]  [%.4f, %.4f]  [%.4f, %.4f]" %
                       (min(r[:,0]), max(r[:,0]), min(r[:,1]), max(r[:,1]),
                        min(r[:,2]), max(r[:,2])))
            if world.size > 1:
                print("Ghost data on processor", i)
                for key in atoms.ghosts.keys():
                    print("  ", key, atoms.ghosts[key].shape)
                r = atoms.ghosts['positions']
                if len(r):
                    print("Limits to the ghost positions:")
                    print ("[%.4f, %.4f]  [%.4f, %.4f]  [%.4f, %.4f]" %
                           (min(r[:,0]), max(r[:,0]), min(r[:,1]), max(r[:,1]),
                            min(r[:,2]), max(r[:,2])))
        world.barrier()


if world.size == 1 or world.rank == 1:
    atoms = FaceCenteredCubic(symbol="Cu", size=(20,20,30))
else:
    atoms = None

if world.size > 1:
    atoms = MakeParallelAtoms(atoms, layout, distribute=False)
    natoms = atoms.get_global_number_of_atoms()
else:
    natoms = len(atoms)
    
report()

print()
print()

if world.size > 1:
    atoms.distribute()

report()

print()
print()
print("Calculating total energy.")
#if world.rank == 0:
#    set_verbose(1)

atoms.calc = EMT()
e = atoms.get_potential_energy()

report()


ReportTest("Potential energy", e, -2.40461833565e-3 * natoms/4, 1e-5)

print("Exiting.")
world.barrier()
#ReportTest.Summary()



"""Test the FixAtoms constraint and the Subset filter."""

from asap3 import EMT, Atoms, Langevin
from ase.lattice.cubic import FaceCenteredCubic
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.md.verlet import VelocityVerlet
from asap3.md.nvtberendsen import NVTBerendsen
from asap3.constraints import Filter
from asap3.constraints import FixAtoms
from ase.constraints import FixAtoms as ASE_FixAtoms
from asap3.test.pytest_markers import ReportTest, serial, parallel
from ase import units
from ase.parallel import world

import numpy as np

import pytest

@serial
@pytest.mark.core
def test_constraints():
    init = FaceCenteredCubic(size=(10,10,10), symbol='Cu', pbc=False)
    z = init.get_positions()[:,2]
    fixedatoms = np.less(z, 0.501*z.max())
    print(len(init), sum(fixedatoms))
    MaxwellBoltzmannDistribution(init, temperature_K=2000)

    print()
    print("Running simulation with Filter")
    atoms1 = Atoms(init)
    atoms1.calc = EMT()
    atoms1a = Filter(atoms1, mask=np.logical_not(fixedatoms))

    dyn = VelocityVerlet(atoms1a, 0.5*units.fs)
    dyn.run(50)
    r1 = atoms1.get_positions()

    print()
    print("Running simulation with Asap's FixAtoms")
    atoms2 = Atoms(init)
    atoms2.calc = EMT()
    atoms2.set_constraint(FixAtoms(mask=fixedatoms))

    dyn = VelocityVerlet(atoms2, 0.5*units.fs)
    dyn.run(50)
    r2 = atoms2.get_positions()

    print()
    print("Running simulation with ASE's FixAtoms")
    atoms3 = Atoms(init)
    atoms3.calc = EMT()
    # Just to be difficult, convert mask to indices
    indx = np.compress(fixedatoms, np.arange(len(atoms3)))
    assert len(indx) == fixedatoms.sum()
    atoms3.set_constraint(ASE_FixAtoms(indx))

    dyn = VelocityVerlet(atoms3, 0.5*units.fs)
    dyn.run(50)
    r3 = atoms2.get_positions()

    err = np.max(np.abs(r1 - r2).flat)
    print()
    print("Filter and Asap's FixAtoms:", err)
    ReportTest("Identical positions (Filter and Asap's FixAtoms)", err, 0.0, 1e-9)

    err = np.max(np.abs(r2 - r3).flat)
    print("ASE's and Asap's FixAtoms:", err)
    ReportTest("Identical positions (ASE's and Asap's FixAtoms)", err, 0.0, 1e-9)

    print()
    print("Running Langevin simulation with Asap's FixAtoms")
    atoms4 = Atoms(init)
    atoms4.calc = EMT()
    atoms4.set_constraint(FixAtoms(mask=fixedatoms))

    dyn = Langevin(atoms4, 0.5*units.fs, temperature_K=1000, friction=0.01)
    dyn.run(50)

    print()
    print("Running Langevin simulation with ASE's FixAtoms")
    atoms5 = Atoms(init)
    atoms5.calc = EMT()
    atoms5.set_constraint(ASE_FixAtoms(mask=fixedatoms))

    dyn = Langevin(atoms5, 0.5*units.fs, temperature_K=1000, friction=0.01)
    dyn.run(50)

    print()
    sanity = [[atoms1, "Verlet + Filter"],
            [atoms2, "Verlet + Asap's FixAtoms"],
            [atoms3, "Verlet + ASE's FixAtoms"],
            [atoms4, "Langevin + Asap's FixAtoms"],
            [atoms5, "Langevin + ASE's FixAtoms"],
            ]
    r_init = init.get_positions()
    for a, label in sanity:
        print(f"Sanity check, {label}:")
        ok = (a.get_positions() == r_init) + np.logical_not(fixedatoms)[:,np.newaxis]
        assert ok.all(), f'Stationary atoms have not moved ({label})'
        ok = (a.get_positions() != r_init) + fixedatoms[:,np.newaxis]
        assert ok.all(), f'Mobile atoms have moved ({label})'


@parallel
def test_par_constraints(cpulayout):
    from asap3 import MakeParallelAtoms

    ismaster = world.rank == 0
    isparallel = world.size != 1
    if ismaster:
        size = 5 * np.array(cpulayout)
        init = FaceCenteredCubic(size=size, symbol='Cu', pbc=False)
        z = init.get_positions()[:,2]
        fixedatoms = np.less(z, 0.501*z.max())
        print(len(init), sum(fixedatoms))
        MaxwellBoltzmannDistribution(init, temperature_K=6000)
        init.set_tags(fixedatoms)
    else:
        init = None

    print()
    print("Running simulation with Filter")
    atoms1 = MakeParallelAtoms(init, cpulayout)
    atoms1.arrays['r_init'] = atoms1.get_positions()
    atoms1.calc = EMT()
    atoms1a = Filter(atoms1, mask=np.logical_not(atoms1.get_tags()))

    dyn = VelocityVerlet(atoms1a, 3*units.fs)
    dyn.run(1000)

    print()
    print("Running simulation with Asap's FixAtoms")
    atoms2 = MakeParallelAtoms(init, cpulayout)
    atoms2.arrays['r_init'] = atoms2.get_positions()
    atoms2.calc = EMT()
    atoms2.set_constraint(FixAtoms(mask=atoms2.get_tags().astype(bool)))

    dyn = VelocityVerlet(atoms2, 3*units.fs)
    dyn.run(1000)

    print()
    print("Running NPTBerendsen simulation with Asap's FixAtoms")
    atoms3 = MakeParallelAtoms(init, cpulayout)
    atoms3.arrays['r_init'] = atoms3.get_positions()
    atoms3.calc = EMT()
    atoms3.set_constraint(FixAtoms(mask=atoms3.get_tags().astype(bool)))

    dyn = NVTBerendsen(atoms3, 3*units.fs, temperature_K=3000, taut=200*units.fs)
    dyn.run(1000)


    print()
    print("Running Langevin simulation with Asap's FixAtoms")
    atoms4 = MakeParallelAtoms(init, cpulayout)
    atoms4.arrays['r_init'] = atoms4.get_positions()
    atoms4.calc = EMT()
    atoms4.set_constraint(FixAtoms(mask=atoms4.get_tags().astype(bool)))

    dyn = Langevin(atoms4, 3*units.fs, temperature_K=3000, friction=0.01)
    dyn.run(1000)

    print()
    print("Running Verlet then Langevin simulation with Asap's FixAtoms")
    atoms5 = MakeParallelAtoms(init, cpulayout)
    atoms5.arrays['r_init'] = atoms5.get_positions()
    atoms5.calc = EMT()
    atoms5.set_constraint(FixAtoms(mask=atoms5.get_tags().astype(bool)))

    dyn = VelocityVerlet(atoms5, 3*units.fs)
    dyn.run(1000)

    dyn = Langevin(atoms5, 3*units.fs, temperature_K=3000, friction=0.01)
    dyn.run(1000)


    print()
    print("Running NVTBerendsen then Verlet simulation with Asap's FixAtoms")
    atoms6 = MakeParallelAtoms(init, cpulayout)
    atoms6.arrays['r_init'] = atoms6.get_positions()
    atoms6.calc = EMT()
    atoms6.set_constraint(FixAtoms(mask=atoms6.get_tags().astype(bool)))

    dyn = NVTBerendsen(atoms6, 3*units.fs, temperature_K=3000, taut=200*units.fs)
    dyn.run(1000)

    dyn = VelocityVerlet(atoms6, 3*units.fs)
    dyn.run(1000)


    print()
    print("Running Verlet then NVTBerendsen simulation with Asap's FixAtoms")
    atoms7 = MakeParallelAtoms(init, cpulayout)
    atoms7.arrays['r_init'] = atoms7.get_positions()
    atoms7.calc = EMT()
    atoms7.set_constraint(FixAtoms(mask=atoms7.get_tags().astype(bool)))

    dyn = VelocityVerlet(atoms7, 3*units.fs)
    dyn.run(1000)

    dyn = NVTBerendsen(atoms7, 3*units.fs, temperature_K=3000, taut=200*units.fs)
    dyn.run(1000)

    print()
    sanity = [[atoms1, "Verlet + Filter"],
              [atoms2, "Verlet + Asap's FixAtoms"],
              [atoms3, "NVTBerendsen + Asap's FixAtoms"],
              [atoms4, "Langevin + Asap's FixAtoms"],
              [atoms5, "Verlet + Langevin + Asap's FixAtoms"],
              [atoms6, "NVTBerendsen + Verlet + Asap's FixAtoms"],
              [atoms7, "Verlet + NVTBerendsen + Asap's FixAtoms"],
             ]
    for a, label in sanity:
        print(world.rank, f"Sanity check, {label}:")
        r_init = a.arrays['r_init']
        ok = (a.get_positions() == r_init) + np.logical_not(a.get_tags())[:,np.newaxis]
        assert ok.all(), 'Stationary atoms have not moved'
        ok = (a.get_positions() != r_init) + a.get_tags()[:,np.newaxis]
        ok.all(), 'Mobile atoms have moved: OK'

        



#! /usr/bin/env python

"""Usage: AsapFileToTrajectory.py oldfile newfile [frame1 [frame2]]

Converts an Asap version 1.x netcdf file to an ASE version 2
trajectory file.  If a single frame number is given, only that frame
is converted.  If two frame numbers are given, all frame from the
first to the last are converted (-1 means the last frame).  If no
frame numbers are given, all frames are converted.

"""

import Numeric as num
from Scientific.IO.NetCDF import NetCDFFile
import sys
import types

old_names = {
    'cartesianPositions': 'CartesianPositions',
    'cartesianMomenta': 'CartesianMomenta',
    'basisVectors': 'UnitCell',
    'classes': 'Tags',
    'atomicNumbers': 'AtomicNumbers',
    'periodic': 'BoundaryConditions'
    }

new_names = {
    #    name                 shape        typecode    once    units
    # -----------------------------------------------------------------
    'CartesianVelocities': (('natoms', 3), num.Float, False, (2, -0.5)),
    'CartesianPositions':  (('natoms', 3), num.Float, False, (1, 0)),
    'CartesianMomenta':    (('natoms', 3), num.Float, False, (2, -0.5)),
    'CartesianForces':     (('natoms', 3), num.Float, False, (-1, 1)),
    'Stress':              ((3, 3),        num.Float, False, (-3, 1)),
    'UnitCell':            ((3, 3),        num.Float, False, (1, 0)),
    'BoundaryConditions':  ((3,),          num.Int,   True , (0, 0)),
    'PotentialEnergy':     ((),            num.Float, False, (0, 1)),
    'AtomicNumbers':       (('natoms',),   num.Int,   True,  (0, 0)),
    'MagneticMoments':     (('natoms',),   num.Float, True,  (0, 0)),
    'Tags':                (('natoms',),   num.Int,   True,  (0, 0))}

def normalize(fr, nfr, default):
    if fr == None:
        fr = default
    if fr < 0:
        fr = nfr + fr
    if fr < 0:
        raise ValueError("Frame number before beginning of file.")
    if fr > nfr:
        raise ValueError("Frame number after end of file: " + str(fr) + " > " + str(nfr))
    return fr

def AsapFileToTrajectory(oldfile, newfile, firstframe=None, lastframe=None):
    # Check if input file is a filename or a NetCDF file
    if isinstance(oldfile, str):
        oldfile = NetCDFFile(oldfile)

    pos = oldfile.variables['cartesianPositions']  # Must be present
    (nframes, natoms, three) = pos.shape
    print(natoms, three, nframes)
    firstframe = normalize(firstframe, nframes, 0)
    lastframe = normalize(lastframe, nframes, -1)
    if lastframe < firstframe:
        raise ValueError("No frames to copy, giving up.")

    print("Preparing to copy frames", firstframe, "to", lastframe)
    # Now open the output file, and define the variables.
    if isinstance(newfile, str):
        newfile = NetCDFFile(newfile, "w")
    oncevars = []
    manyvars = []
    for v in oldfile.variables.keys():
        try:
            newname = old_names[v]
        except KeyError:
            print("WARNING: Skipping data named", v)
            continue
        if new_names[newname][2]:
            shape = new_names[newname][0]
            oncevars.append((v, newname))
        else:
            shape = ("unlim",) + new_names[newname][0]
            manyvars.append((v, newname))
        shape2 = []
        for d in shape:
            if isinstance(d, int):
                n = d
                d = str(d)
            elif d == 'natoms':
                n = natoms
            elif d == 'unlim':
                n = None
            else:
                raise RuntimeError("Unknown dimension "+str(d))
            if d not in newfile.dimensions:
                newfile.createDimension(d, n)
            shape2.append(d)
        print(v, "-->", newname, " shape", shape2)
        var = newfile.createVariable(newname, oldfile.variables[v].typecode(),
                                     tuple(shape2))
        var.once = new_names[newname][2]
        var.units = new_names[newname][3]
        
    # Now copy the data
    print("Copying global data")
    newfile.history = 'ASE trajectory'
    newfile.version = '0.1'
    newfile.lengthunit = 'Ang'
    newfile.energyunit = 'eV'
    for oldname, newname in oncevars:
        newfile.variables[newname][:] = oldfile.variables[oldname][:]
    
    for n in range(firstframe, lastframe+1):
        print("Copying frame", n)
        for oldname, newname in manyvars:
            newfile.variables[newname][n] = oldfile.variables[oldname][n]
    newfile.close()
    
if __name__ == "__main__":
    # sys.tracebacklimit = 0
    if len(sys.argv) < 3 or len(sys.argv) > 5:
        print(__doc__)
        raise TypeError("Wrong number of arguments.")
    infile = sys.argv[1]
    outfile = sys.argv[2]
    first = last = None
    if len(sys.argv) > 3:
        first = last = int(sys.argv[3])
    if len(sys.argv) > 4:
        last = int(sys.argv[4])
    AsapFileToTrajectory(infile, outfile, first, last)
    



# Test that various observers and analysis tools do not interfere with
# the energy calculator e.g. by messing up the neighbor list.
#
# It works by making a short simulation without any observers, and then
# repeating it with observers, checking that the result is the same.

from asap3 import EMT, Atoms, units, MakeParallelAtoms
from asap3.md.verlet import VelocityVerlet
from asap3.md.langevin import Langevin
from ase.lattice.cubic import FaceCenteredCubic
from asap3.analysis import RestrictedCNA, RadialDistributionFunction
from asap3.test.pytest_markers import ReportTest, serial
from ase.parallel import world
import numpy as np

import pytest

def Compare(name, atoms, observer, ref, interval=4):
    dyn = VelocityVerlet(atoms, 5*units.fs)
    dyn.attach(observer, interval=interval)
    r = []
    for i in range(50):
        dyn.run(5)
        epot = atoms.get_potential_energy() / len(atoms)
        ekin = atoms.get_kinetic_energy() / len(atoms)
        r.append([epot, ekin, epot+ekin])
    r = np.array(r)
    diff = r - ref
    maxdiff = diff.max()
    print("Maximal difference is", maxdiff)
    ReportTest(name, maxdiff, 0.0, 1e-6)
    
@serial   # It looks like RadialDistributionFunction coredumps in parallel
@pytest.mark.core
def test_nointerference(cpulayout):
    if cpulayout:
        size = np.array((5, 5, 5)) * cpulayout
    else:
        size = (10, 5, 5)
    print("Making initial system")
    if world.rank == 0:
        iniatoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                                    size=size, symbol="Cu", pbc=(1,1,0))
        iniatoms.calc = EMT()
        inidyn = Langevin(iniatoms, 5*units.fs, temperature_K=450, friction=0.05)
        inidyn.run(100)
        print("Temperature is now", iniatoms.get_kinetic_energy() / (1.5*units.kB*len(iniatoms)), "K")
    else:
        iniatoms = None
    world.barrier()

    print("Making reference simulation")
    if world.size == 1:
        refatoms = Atoms(iniatoms)
    else:
        refatoms = MakeParallelAtoms(iniatoms, cpulayout)
    refatoms.calc = EMT()
    dyn = VelocityVerlet(refatoms, 5*units.fs)
    ref = []
    for i in range(50):
        dyn.run(5)
        epot = refatoms.get_potential_energy() / len(refatoms)
        ekin = refatoms.get_kinetic_energy() / len(refatoms)
        ref.append([epot, ekin, epot+ekin])

    ref = np.array(ref)

    print("Testing RestrictedCNA")
    if world.size == 1:
        atoms = Atoms(iniatoms)
    else:
        atoms = MakeParallelAtoms(iniatoms, cpulayout)
    atoms.calc = EMT()
    cna = RestrictedCNA(atoms)
    Compare("RestrictedCNA", atoms, cna.analyze, ref)

    print("Testing RadialDistributionFunction")
    if world.size == 1:
        atoms = Atoms(iniatoms)
    else:
        atoms = MakeParallelAtoms(iniatoms, cpulayout)
    atoms.calc = EMT()
    rdf = RadialDistributionFunction(atoms, 4.0, 25)
    Compare("RadialDistributionFunction", atoms, rdf.update, ref)




from asap3 import NeighborCellLocator
from asap3.setup.multidislocation import random_dislocations
from asap3.setup.dislocation import Dislocation
from asap3.memory import get_max_memory
from ase.atoms import Atoms
from ase.lattice.cubic import FaceCenteredCubic, BodyCenteredCubic
import ase.data
import numpy as np
import scipy.spatial
from multiprocessing import Pool, cpu_count
from . import plane_utils
from . import quat_utils
import math
import sys
import os

def make_nanocrystal(size, centres, rotations, min_dist=2.0, unit=None,
                     symbol=None, latticeconstant=None):
    """Create a nanocrystalline sample.

    Parameters:
      size:      Size of the system, in Angstrom (either a number or three numbers).
      centres:   Positions of grains in scaled coordinates.
      rotations: Rotation matrices for the grains.

      min_dist (optional):
                 If two atoms are closer than this distance, one of them is removed
                 (default: 2.0).

    In addition, either the 'unit' or the 'symbol' parameter must be specified.
      unit (optional):
                 A unit cell for building the crystals (Atoms object).
                 MUST be orthorhombic!
      symbol (optional):
                 If unit is None, then an FCC crystal of this element is used.
      latticeconstant (optional):
                 If symbol is specified, this overrides the default lattice constant.
    """

    assert type(size) == float or type(size) == list
    if type(size) == float:
        size = np.array([size, size, size])
    elif type(size) == list:
        size = np.array(size)
    assert size.shape == (3,)
    assert len(centres) == len(rotations)
    centres *= size
    if unit is None:
        unit = FaceCenteredCubic(symbol=symbol, size=(1,1,1), pbc=True,
                                 latticeconstant=latticeconstant)

    return create_atoms(centres, rotations, unit, size, min_dist)

def make_dislocated_nanocrystal(size, centres, rotations, 
                                dislocationdensity,
                                symbol, latticeconstant=None,
                                min_dist=2.0,
                                multiproc=False):
    """Create a nanocrystalline sample with pre-existing dislocation density.

    Parameters:
      size:      Size of the system, in Angstrom (either a number or three numbers).
      centres:   Positions of grains in scaled coordinates.
      rotations: Rotation matrices for the grains.
      dislocationdensity:
                 Dislocation density in m^-2
      symbol:    Chemical symbol or atomic number.
      latticeconstant (optional):
                 Overrides the default lattice constant.
      min_dist (optional):
                 If two atoms are closer than this distance, one of them is removed
                 (default: 2.0).
      multiproc: Use all cores to calculate dislocation strain fields (default: False).
    """

    assert type(size) == float or type(size) == list
    if type(size) == float:
        size = np.array([size, size, size])
    elif type(size) == list:
        size = np.array(size)
    assert size.shape == (3,)
    assert len(centres) == len(rotations)
    centres *= size
    if latticeconstant is None:
        # Get lattice constant from the element
        if isinstance(symbol, int):
            z = symbol
        else:
            z = ase.data.atomic_numbers[symbol]
        if ase.data.reference_states[z]['symmetry'] != 'fcc':
            raise ValueError(f"Cannot extract FCC lattice constant from non-fcc element ({symbol})")
        latticeconstant = ase.data.reference_states[z]['a']
    unit = FaceCenteredCubic(symbol=symbol, size=(1,1,1), pbc=True,
                             latticeconstant=latticeconstant)
    # Create a dislocation field for each grain.  The radius of the dislocation
    # field should be half the sidelength of the overall system
    dislocationfields = []
    for i in range(len(centres)):
        disloc_params = random_dislocations(max(size)/2.0, dislocationdensity)
        n_dislocs = len(disloc_params)
        dislocs = [Dislocation(point, line, latticeconstant * burgers)
                   for point, burgers, line in disloc_params]
        dislocationfields.append(dislocs)

    return create_atoms(centres, rotations, unit, size, min_dist,
                        dislocations=dislocationfields, multiproc=multiproc)


# A function for generating random rotation matrices
random_rotation_matrix = quat_utils.random_rotation_matrix


def bcc_grains(size, perturbation=0.0):
    "Creates a grain layout based on a perturbed BCC lattice."
    graincenters = BodyCenteredCubic(symbol='H', latticeconstant=1.0, size=size)
    graincenters = graincenters.get_positions() + 0.25
    graincenters /= size
    pert = np.random.standard_normal(graincenters.shape) * perturbation
    graincenters += pert
    rotations = [random_rotation_matrix() for g in graincenters]
    return graincenters, rotations

def minimize_energy(atoms, nstep, pressure_interval=10, bulkmodulus = 140e9/1e5):
    """Helper function for minimizing the energy of a nanocrystalline structure.

    It minimizes the energy while eliminating the diagonal components of the stress.
    The atomic positions are optimized using the FIRE algorithm and the stress
    is minimized using the Inhomogeneous_Berendsen algorithm.

    Parameters:

    atoms: The atoms object to be energy-minimized.

    nsteps: Number of time steps in the FIRE algorithm

    pressure_interval=10: (optional) How often to call the stress minimizer.

    bulkmodulus=1.4e6: (optional).  The bulk modulus (in bar !!!) used by the
    stress optimizer.  This value is for Cu and is useful for all metals.
    The value is uncritical, but the order of magnitude should be reasonable. 
    """
    dyn = FIRE(atoms)
    unstress = Inhomogeneous_NPTBerendsen(atoms, 50*units.fs, 0,
                                          taup=5000*units.fs,
                                          compressibility=1/bulkmodulus)
    dyn.attach(unstress.scale_positions_and_cell, interval=10)
    dyn.run(steps=nstep)


###
###  Below are helper functions for make_nanocrystalline
###


def remove_close(atoms, min_dist):
    "Remove atoms that are too close"
    print(f"Removing atoms that are too close (d_cut = {min_dist:.3f} Angstrom)")
    remove = set()
    nblist = NeighborCellLocator(min_dist, atoms)
    # Loop over all pairs of too-close atoms
    for i in range(len(atoms)):
        for j in nblist[i]:
            # Pair (i,j) is too close
            if not (i in remove or j in remove):
                if len(remove) % 2:
                    remove.add(i)
                else:
                    remove.add(j)
    print("Removing %i atoms out of %i" % (len(remove), len(atoms)))
    del atoms[list(remove)]
    return atoms

def calculate_grain_bounding_planes(centres, size):

    a = (-1, 0, 1)
    ctr = [c for c in centres]
    for i in a:
        for j in a:
            for k in a:
                if i == j == k == 0: continue
                ctr += [size * [i, j, k] + c for c in centres]
    ctr = np.array(ctr)

    voronoi = scipy.spatial.Voronoi(ctr, furthest_site=False,
                                    incremental=False, qhull_options=None)
    regions = [voronoi.regions[voronoi.point_region[i]] for i in range(len(centres))]

    grain_cells = []
    for i, reg in enumerate(regions):

        cell_vertices = np.array([voronoi.vertices[v] for v in reg])
        hull = scipy.spatial.ConvexHull(cell_vertices)

        bounding_planes = [plane_utils.plane_equation(centres[i], h, cell_vertices)
                           for h in hull.simplices]
        grain_cells += [(cell_vertices, bounding_planes)]

    return grain_cells

_field_to_apply = None

def apply_fields(r):
    global _field_to_apply
    res = np.zeros_like(r)
    for f in _field_to_apply:
        res += f(r)
    return res

_positions_args = None
def create_layer_positions(i):
    global _positions_args
    y0, y1, z0, z1, unit = _positions_args
    positions = [np.dot((i, j, k), unit.cell) + atom_pos
                 for j in range(y0, y1+1)
                 for k in range(z0, z1+1)
                 for atom_pos in unit.get_positions()]
    return np.array(positions)

def create_atoms(centres, rotations, unit, size, min_dist, dislocations=None,
                 multiproc=False):
    "Create the nanocrystalline sample"
    global _field_to_apply, _positions_args

    if multiproc:
        if multiproc is True:
            multiproc_cpus = None  # Use default
        else:
            if not isinstance(multiproc, int) or multiproc <= 0:
                raise TypeError("multiproc argument must be a positive integer.")
            multiproc_cpus = multiproc  # Use specified number
            
    grain_cells = calculate_grain_bounding_planes(centres, size)

    ntotal = 0
    result = None
    for grain_num, (centre, rotation) in enumerate(zip(centres, rotations)):
        assert centre.shape == (3,)
        assert rotation.shape == (3,3)

        (bounding_vertices, bounding_planes) = grain_cells[grain_num]

        #find maximum extent in [x, y, z]
        U = rotation
        rot_basis = U / np.linalg.norm(U, axis=0)

        proj_vertices = np.dot(bounding_vertices, rot_basis)

        xmin = np.min(proj_vertices[:,0])
        ymin = np.min(proj_vertices[:,1])
        zmin = np.min(proj_vertices[:,2])

        xmax = np.max(proj_vertices[:,0])
        ymax = np.max(proj_vertices[:,1])
        zmax = np.max(proj_vertices[:,2])

        #calculate number of unit cells in each direction
        cubic_cell = [unit.cell[0][0], unit.cell[1][1], unit.cell[2][2]]

        x0 = int(math.floor(xmin / cubic_cell[0]))
        y0 = int(math.floor(ymin / cubic_cell[1]))
        z0 = int(math.floor(zmin / cubic_cell[2]))

        x1 = int(math.ceil(xmax / cubic_cell[0]))
        y1 = int(math.ceil(ymax / cubic_cell[1]))
        z1 = int(math.ceil(zmax / cubic_cell[2]))

        # If there are dislocations, make space for their worst-case
        # cumulative dislplacement
        if dislocations:
            field = dislocations[grain_num]
            n_disl = len(field)
            correction = int(n_disl // np.sqrt(2))
            x0 -= correction
            y0 -= correction
            z0 -= correction
            x1 += correction
            y1 += correction
            z1 += correction
            
        #create cube of atoms    #todo: outer product
        if multiproc:
            _positions_args = (y0, y1, z0, z1, unit)
            pool = Pool(multiproc_cpus)
            positions = pool.map(create_layer_positions, range(x0, x1+1))
            positions = np.concatenate(positions)
            pool.close()
        else:
            positions = [np.dot((i, j, k), unit.cell) + atom_pos
                         for i in range(x0, x1+1)
                         for j in range(y0, y1+1)
                         for k in range(z0, z1+1)
                         for atom_pos in unit.get_positions()]
            positions = np.array(positions)
        print("Temporary length of position array:", len(positions))
        sys.stdout.flush()
        # Apply dislocation field, if any
        if dislocations:
            centre_of_mass = positions.sum(axis=0)/len(positions)
            if multiproc:
                if multiproc_cpus:
                    ncores = multiproc_cpus
                else:
                    ncores = cpu_count()
                print(f"Using {ncores} cores.")
                _field_to_apply = field  # Cannot pass a lambda to Pool.map
                arg = np.array_split(positions - centre_of_mass, ncores)
                pool = Pool(multiproc_cpus)
                res = pool.map(apply_fields, arg)
                totalfield = np.concatenate(res)
                pool.close()
                assert totalfield.shape == positions.shape
            else:
                totalfield = np.zeros_like(positions)
                for i, f in enumerate(field):
                    totalfield += f(positions - centre_of_mass)
            positions = positions + totalfield
            print(n_disl, "dislocations introduced.")
            sys.stdout.flush()
        # Apply rotation
        positions = np.dot(positions, U.transpose())
        
        #remove excess atoms
        print("Removing excess atoms.")
        in_cell = np.ones(len(positions)).astype(np.int8)
        for (p, n) in bounding_planes:
            in_cell = np.logical_and(in_cell, np.dot(n, (positions - p).T) > 0)

        positions = positions[in_cell]

        print("Creating Atoms object, and merging...")
        atoms = Atoms(positions = positions, cell=size, numbers = 29 * np.ones(len(positions)))
        atoms.set_tags(grain_num * np.ones(len(atoms), int))
        if result is None:
            result = atoms
            atoms.set_cell(size, scale_atoms=False)
            atoms.set_pbc(True)
        else:
            result.extend(atoms)
        print("grain %d contains %d atoms (max mem. %.1f GB)."
              % (grain_num, len(positions), get_max_memory()/1024))
        sys.stdout.flush()

    print("Total atoms:", len(result))
    result = remove_close(result, min_dist)
    print("Remaining atoms:", len(result))
    return result




import sys
import time
import signal
import numpy as np

from exceptions import KeyboardInterrupt

from asap3.Tools.ParameterOptimization.Quantities import QuantityCalc
from asap3.Tools.ParameterOptimization.ScipyFmin import fmin, fmin_bfgs, fmin_powell
from asap3.Tools.Timing import Timer
from ase.utils import basestring
import numbers

class TimeoutException(Exception):
    pass

def timeout_handler(signum, frame):
    raise TimeoutException("Calculation exceeded the time limit!")

class ParameterOptimization():
    """Optimizes potential parameters based on given quantities.

    Input format of the potential parameters:

    {(elements): [parameters], ...}

    Input format of the quantities is:

    [[name, structure, elements, value, deviation], ...]

    The following quantities and structures are supported:

    Name                    Structures
    ----                    ----------
    cohesive_energy         fcc, hcp, bcc
    vacancy_energy          fcc, hcp, bcc
    heat_of_formation       l12, l10, B2
    phase_energy            fcc-hcp, fcc-bcc, bcc-hcp
    lattice_constant_a      fcc, hcp, l12, l10, B2
    lattice_constant_c      hcp, l10
    lattice_constant_ca     hcp, l10
    bulk_modulus            fcc, hcp, bcc, l12, B2
    elastic_constant_C11    fcc, hcp, bcc, l12, B2
    elastic_constant_C12    fcc, hcp, bcc, l12, B2
    elastic_constant_C13    hcp
    elastic_constant_C33    hcp
    elastic_constant_C44    fcc, hcp, bcc, l12, B2
    surface_energy          fcc100, fcc111, hcp0001
    surface_ratio           any combination of surface energies
                            on the form fcc100-fcc111
    impurity_energy         oct38-face, oct38-edge

    """

    def __init__(self, elements, calc_class, parameters, varparameters, quantities,
                 latticeconstants, calc_args=None, debug=False):
        self.calc_class = calc_class
        if isinstance(calc_args, tuple):
            self.calc_args = calc_args
        else:
            self.calc_args = ()

        self.debug = debug
        self.timer = Timer()
        self.count_errors = 0
          
        # Interpret elements
        if not isinstance(elements, (tuple, list)):
            elements = [elements]
        #assert len(elements) < 3
        self.elements = list(elements)

        # Interpret parameters
        self.param_keys = []
        self.param_value = []
        self.param_variable = []
        for key in parameters.keys():
            if isinstance(key, tuple) and len(key) == 2:
                for symbol in key:
                    if symbol not in self.elements:
                        raise ValueError("Symbol in key must be in the element list")
                self.param_keys.append(key)
                self.param_value.append(parameters[key][:])
                self.param_variable.append(varparameters[key][:])
            else:
                raise KeyError("Parameter symbol key must be a tuple of lenght two.")

        # Interpret quantities
        self.quan_calc = QuantityCalc(latticeconstants, None, self.timer, self.debug)
        self.quan_name = []
        self.quan_structure = []
        self.quan_elements = []
        self.quan_value = []
        self.quan_deviation = []
        for quan in quantities:
            if not hasattr(self.quan_calc, 'get_' + quan[0]):
                raise ValueError(f"The '{quan[0]}' quantity cannot be calculated.")
            self.quan_name.append(quan[0])
            self.quan_structure.append(quan[1])
            if isinstance(quan[2], (basestring, numbers.Integral)):
                self.quan_elements.append(tuple([quan[2]]))
            else:
                self.quan_elements.append(tuple(quan[2]))
            self.quan_value.append(quan[3])
            self.quan_deviation.append(quan[4])

    def valuelist_to_quantities(self, values, error=None):
        quantities = []
        for name, structure, elements, value in zip(self.quan_name,
                                                    self.quan_structure,
                                                    self.quan_elements,
                                                    values):
            key = (name, structure, elements)
            if error != None and error >= 1e9 - 2:
                value = 0.0
            quantities.append([name, structure, elements, value])

        return quantities

    def quantities_to_valuelist(self, quantities):
        values = []
        for i, quan in enumerate(quantities):
            assert self.quan_name[i] == quan[0]
            values.append(quan[3])
        return values

    def parameters_list2dict(self, paramlist, copyall=False):
        """Copies a list of parameters (used in optimization) to a dictionary
        (used in potentials). Copies by default only variable parameters - set
        copyall=True to copy all parameters."""
        k = 0
        paramdict = {}
        for i, key in enumerate(self.param_keys):
            paramdict[key] = self.param_value[i][:]
            for j, bool in enumerate(self.param_variable[i]):
                if bool or copyall:
                    paramdict[key][j] = paramlist[k]
                    k += 1
        return paramdict

    def parameters_dict2list(self, paramdict=None, copyall=False):
        """Reverse of parameters_list2dict."""
        paramlist = []
        for i, key in enumerate(self.param_keys):
            for j, bool in enumerate(self.param_variable[i]):
                if bool or copyall:
                    if paramdict == None:
                        paramlist.append(self.param_value[i][j])
                    else:
                        paramlist.append(paramdict[key][j])
        return paramlist

    def fit(self, parameters=None, xtol=1e-2, ftol=5e-2, delta=5e-3,
            log=None, err=None, rawout=False):
        # Open and print log
        if log is not None:
            if isinstance(log, basestring):
                self.log = open(log, 'a')
            else:
                # Hopefully an open file.
                self.log = log
            #assert hasattr(self.log, 'write')
            
            output = "Time      Error Function    "
            format = "%02i:%02i:%02i  %-18.8e"
            T = time.localtime()
            strargs = [T[3], T[4], T[5], 0.0]
            for i, key in enumerate(self.param_keys):
                for j, bool in enumerate(self.param_variable[i]):
                    if bool:
                        output += "P%i.%-8i" % (i,j)
                        format += "%-11.6f"
                        strargs.append(0.0)
            for i, value in enumerate(self.quan_value):
                output += "Q%-9s" % (i,)
                format += "%-10.3f"
                strargs.append(value)
            T = time.localtime()
            self.log.write("\n" + output + "\n")
            self.log.write(len(output) * "-" + "\n")
            self.log.write(format % tuple(strargs) + "\n")
            self.log.flush()
            self.log_format = format
        else:
            self.log = None

        # Open error file/pipe
        if err is not None:
            if isinstance(err, basestring):
                self.err = open(err, 'w')
            else:
                # Hopefully an open file.
                self.err = err
            #assert hasattr(self.err, 'write')
        else:
            self.err = None

        # Copy the variable parameters to a list - if its is a list assume 
        # only varible parameters
        if not isinstance(parameters, list):
            parameters = self.parameters_dict2list(parameters)

        # Start optimization
        xopt, fopt, direc, iter, calls, flag = fmin_powell(self.error_function, parameters,
                                                           xtol=xtol, ftol=ftol,
                                                           full_output=True, disp=False)

        # Print log after optimization
        if self.log is not None:
            if flag == 0:
                self.log.write("\n\nFitting terminated successfully.\n")
            else:
                self.log.write("\n\nFitting not converged - " +
                                "too many iterations or function calls.\n")
            self.log.write(f"   Error function value: {fopt:f}\n" +
                            "   Iterations: %d\n" % (iter,) +
                            "   Error function evaluations: %d\n\n\n" % (calls,))
            #self.timer.write(self.log)
        else:
            if flag > 0:
                print ("Fitting not converged - " +
                       "too many iterations or function calls!")

        if fopt > 1e99:
            self.count_errors += 1
            if self.count_errors >= 50:
                raise RuntimeError("50 optimizations failed in a row.  Aborting.")
        else:
            self.count_errors = 0
        # Replace the old parameters with the optimized ones
        parameters = self.parameters_list2dict(xopt)

        # Get last calculated values of quantities
        (error, values) = self.calc_values(parameters)
        quantities = self.valuelist_to_quantities(values, error)

        if rawout:
            return (error, list(xopt), values)
        else:
            return (error, parameters, quantities)

    def calc_values(self, parameters):
        # Set calculator and calculate values
        self.timer.start('calc_quantity_values')
        try:
            calc = self.calc_class(self.elements[:], parameters, *self.calc_args)
            self.quan_calc.set_calculator(calc)
        except:
            T = time.localtime()
            self.err.write("Calculation Error..." +
                           "\n   Time: %02i-%02i-%i" % (T[2], T[1], T[0]) +
                           " %02i:%02i:%02i" % (T[3], T[4], T[5]) +
                           f"\n   Type: {sys.exc_info()[0]}" + 
                           f"\n   Message: {sys.exc_info()[1]}" + 
                           f"\n   Parameters: {parameters}" + "\n\n")
            self.err.flush()
            values = [0.0] * len(self.quan_name)
            error = 2e100 - 1
        else:
            error = 0.0
            values = []
            for name, structure, elements, fitvalue, fitdev in zip(self.quan_name,
                                                                   self.quan_structure,
                                                                   self.quan_elements,
                                                                   self.quan_value,
                                                                   self.quan_deviation):
                old_handler = signal.signal(signal.SIGALRM, timeout_handler)
                if 'surface' in name:
                    signal.alarm(20)
                else:
                    signal.alarm(4)
                # Triger an alarm in 2 seconds. Simpel statistics show that the
                # mean calculation time for a quantity lies around 0.5 seconds,
                # with a maximun registred around 1 second.

                try:
                    try:
                        value = self.quan_calc.get_quantity(name, structure, elements)
                        if isinstance(value, complex):
                            raise ValueError("Got a complex value.")
                        if fitvalue == 0:
                            error += (value - fitvalue)**2 / (fitdev)**2
                        else:
                            error += (value - fitvalue)**2 / (fitvalue * fitdev)**2
                    except KeyboardInterrupt:
                        raise
                    except: #(TimeoutException, RuntimeError, AsapError):
                        T = time.localtime()
                        self.err.write("Calculation Error..." +
                                       "\n   Time: %02i-%02i-%i" % (T[2], T[1], T[0]) +
                                       " %02i:%02i:%02i" % (T[3], T[4], T[5]) +
                                       f"\n   Type: {sys.exc_info()[0]}" + 
                                       f"\n   Message: {sys.exc_info()[1]}" + 
                                       f"\n\n   Value: {name}" +
                                       f"\n   Structure: {structure}" +
                                       f"\n   Elements: {elements}" +
                                       f"\n   Parameters: {parameters}" +
                                       f"\n   Values: {values}" + "\n\n")
                        self.err.flush()
                        values = [0.0] * len(self.quan_name)
                        error = 1e100 - 1
                        break
                    else:
                        values.append(value)
                    signal.alarm(0)
                except TimeoutException:
                    pass
                finally:
                    signal.alarm(0)
                    signal.signal(signal.SIGALRM, old_handler)

        self.timer.stop('calc_quantity_values')
        return (error, values)

    def error_function(self, newparameters):
        # Replace the old parameters with the new ones
        parameters = self.parameters_list2dict(newparameters)

        # Calculate values and errorfunction
        (errfunc, values) = self.calc_values(parameters)

        # Print log
        if self.log is not None:
            #if errfunc > 1e9 - 1:
            #    errfunc = 1e9 - 1
            T = time.localtime()
            strargs = [T[3], T[4], T[5], errfunc]
            strargs += newparameters.tolist()
            strargs += values
            try:
                self.log.write(self.log_format % tuple(strargs) + "\n")
            except TypeError:
                sys.stderr.write("Failing to write to log file (TypeError in formatting)\n")
                sys.stderr.write("Format string: "+self.log_format+'\n')
                sys.stderr.write("Arguments: "+str(strargs)+'\n')
            self.log.flush()

        return errfunc

    def write_result(self, fitpar, optpar, optval, out=sys.stdout):
        if isinstance(fitpar, list):
            fitpar = self.parameters_list2dict(fitpar)
        if isinstance(optpar, list):
            optpar = self.parameters_list2dict(optpar)
        if isinstance(optval[0], (tuple, list)):
            optval = self.quantities_to_valuelist(optval)

        for i, key in enumerate(self.param_keys):
            out.write(f"  {key[0]}{key[1]} parameters " +
                      "(Initial, Optimal, Variable):\n")
            for pi, po, v in zip(fitpar[key], optpar[key],
                                 self.param_variable[i]):
                out.write(f"  {pi:12.9f}  {po:12.9f}  {v}\n")
        out.write("\n  Fitting values (Quantity, Structure: " +
                  "Target, Optimal, Weight, Deviation):\n")
        devsum = 0
        for i, name in enumerate(self.quan_name):
            elements = self.quan_elements[i]
            structure = self.quan_structure[i]
            value = self.quan_value[i]
            weight = self.quan_deviation[i]

            if len(elements) > 1:
                structure = "%s%s-%s" % (elements + (structure,))
            else:
                structure = "%s-%s" % (elements + (structure,))
            name = f"{name}, {structure}:"
            if value == 0:
                dev = np.abs(optval[i]) * 100
            else:
                dev = np.abs(optval[i] / value - 1) * 100
            devsum += dev
            out.write("  %-45s  %7.3f  " % (name, value) +
                      f"{optval[i]:10.6f}  {weight:5.3f}  " +
                      f"{dev:7.2f}%\n")
        out.write(f"{devsum / len(self.quan_name):84.2f}%\n\n")

if __name__ == '__main__':
    from asap3.Tools.ParameterOptimization.EMT import EMT2011Fit, EMTStdParameters

    # Parameter names: ['eta2', 'lambda', 'kappa', 'E0', 'V0', 'S0', 'n0']
    initparam = {('Pt','Pt'): EMTStdParameters('Pt')}
    varparam = {('Pt','Pt'): [True, True, True, True, True, True, False]}

    quantities = [('lattice_constant_a', 'fcc', 'Pt', 3.92,      0.001),
                  #('lattice_constant_a', 'hcp', 'Pt', 2.77,      0.05),
                  #('lattice_ratio_ca', 'hcp', 'Pt', 4.78 / 2.77, 0.05),
                  ('bulk_modulus', 'fcc', 'Pt', 278.3,           0.01),
                  #('elastic_anisotropy', 'fcc', 'Pt', 1.594,   0.01),
                  #('elastic_constant_C11', 'fcc', 'Pt', 346.7,   0.1),
                  #('elastic_constant_C12', 'fcc', 'Pt', 250.7,  0.1),
                  #('elastic_constant_C44', 'fcc', 'Pt', 76.5,    0.01),
                  ('cohesive_energy', 'fcc', 'Pt', 5.84,         0.001),
                  #('phase_energy', 'fcc-hcp', 'Pt', -0.05,       0.02),
                  #('surface_energy', 'fcc111', 'Pt', 0.631,      0.02),
                  #('surface_ratio', 'fcc111-fcc100', 'Pt', 0.631 / 0.892, 0.01),
                  ]

    latticeconstants = [['fcc', 'Pt', 3.92],
                        ['hcp', 'Pt', (2.77, 4.78)]]

    opt = ParameterOptimization(['Pt'], EMT2011Fit, initparam, varparam,
                                quantities, latticeconstants, ('kappa',), False)
    (error, optpar, optval) = opt.fit(xtol=0.1, ftol=10.0, delta=0.01,
                                      log=sys.stdout, err='Optimization.err')
    print(f'Optimization ended with error function {error:.6e}')
    opt.write_result(initparam, optpar, optval)




from asap3 import EMT
from asap3.analysis import PTM
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
import ase.data
from asap3.test.pytest_markers import serial

import pytest


@serial
@pytest.mark.core
def test_ptm_small():
    element = "Cu"
    latconst = ase.data.reference_states[ase.data.atomic_numbers[element]]['a']
    nndist = latconst / np.sqrt(2)

    sizes = [(3,3,3), (2, 2, 2), (1, 1, 1)]
    withpotential =  (False, True)
    quicklist = (False, True)

    for size in sizes:
        for usepot in withpotential:
            for quick in quicklist:
                print(f'Size: {size}  -  use potential: {usepot}  -  quick: {quick}.')
                atoms = FaceCenteredCubic(
                    directions=[[1,0,0],[0,1,1],[0,0,1]], 
                    size=size,
                    symbol=element,
                    pbc=True
                )
                
                if usepot:
                    atoms.calc = EMT()
                    epot = atoms.get_potential_energy()

                cutoff = nndist * (1 + np.sqrt(2)) / 2
                ptmdata = PTM(atoms, cutoff=cutoff, rmsd_max=0.1, quick=quick)
                structure = ptmdata['structure']
                assert np.equal(structure, 1).all()





from ase import units
from ase.build import bulk
from asap3 import EMT
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.md.velocitydistribution import Stationary
from asap3.md.langevin import Langevin
from numpy.linalg import norm
from asap3.test.pytest_markers import serial
import numpy as np

import pytest

# parameters
size = 2
T = 300
dt = 0.01

# setup
@pytest.fixture
def atoms():
    atoms = bulk('CuAg', 'rocksalt', a=4.0).repeat(size)
    atoms.pbc = False
    atoms.calc = EMT()
    return atoms

@serial
@pytest.mark.core
@pytest.mark.parametrize('forcease', (False, True))
def test_langevin_com(forcease, atoms):
    MaxwellBoltzmannDistribution(atoms, temperature_K=T)
    Stationary(atoms)
    dyn = Langevin(atoms, dt * units.fs, temperature_K=T, friction=0.02, forceASE=forcease)
    
    # run NVT
    mtot = atoms.get_momenta().sum(axis=0)

    print('initial momenta', mtot)
    print('initial forces', atoms.get_forces().sum(axis=0))

    dyn.run(10)

    m2 = atoms.get_momenta().sum(axis=0)
    print('momenta', m2)
    print('forces', atoms.get_forces().sum(axis=0))
    print()

    assert norm(m2) < 1e-8, f'Langevin changed center of mass with forceASE={forcease}'



from ase.units import fs
from asap3 import EMT
from asap3.md.langevin import Langevin
from ase.build import fcc111
from asap3.test.pytest_markers import serial
import numpy as np

import pytest

@serial
@pytest.mark.parametrize('seed', [42, None])
def test_langevin_seed(seed):
    positions = []
    for _ in range(2):
        atoms = fcc111('Au', size=(3,3,3), vacuum=10.0)
        atoms.calc = EMT()

        dyn = Langevin(atoms, timestep=5.0*fs, temperature_K=300, friction=1e-1, seed=seed)
        dyn.run(1000)

        positions.append(atoms.get_positions())

    p1, p2 = positions
    is_pos_equal = np.allclose(p1, p2)
    print("Max difference:", abs(p1 - p2).max())
    if seed is None:
        assert not is_pos_equal, "Runs differ when not seeded"
    else:
        assert is_pos_equal, "Runs are equal when seeded"



from ase.lattice.cubic import FaceCenteredCubic
from asap3 import FullNeighborList, print_version
from asap3.test.pytest_markers import ReportTest, serial
import ase
import numpy as np

import pytest

class AseNeigborListWrapper:
    """
    Wrapper around ASE neighborlist to have the same interface as asap3 neighborlist
    """

    def __init__(self, cutoff, atoms):
        self.neighborlist = ase.neighborlist.NewPrimitiveNeighborList(
            cutoff, skin=0.0, self_interaction=False, bothways=True
        )
        self.neighborlist.build(
            atoms.get_pbc(), atoms.get_cell(), atoms.get_positions()
        )
        self.cutoff = cutoff
        self.atoms_positions = atoms.get_positions()
        self.atoms_cell = atoms.get_cell()

    def get_neighbors(self, i, cutoff):
        assert (
            cutoff == self.cutoff
        ), "Cutoff must be the same as used to initialise the neighborlist"

        indices, offsets = self.neighborlist.get_neighbors(i)

        rel_positions = (
            self.atoms_positions[indices]
            + offsets @ self.atoms_cell
            - self.atoms_positions[i][None]
        )

        dist2 = np.sum(np.square(rel_positions), axis=1)

        return indices, rel_positions, dist2


def compare_with_ase_neighborlist(atoms, query_points, cutoff):
    """Test FullNeighborList.get_neighbors_querypoint
    by comparing the output with an equivalent implementation in ASE
    that inserts query atoms into an atoms object."""
    # Insert query atoms into ASE atoms object
    query_atoms = ase.Atoms(numbers=[0] * query_points.shape[0], positions=query_points)
    atoms_with_query = atoms.copy()
    atoms_with_query.extend(query_atoms)

    # Compute neighborlists
    asefullneighbors = AseNeigborListWrapper(cutoff, atoms_with_query)
    asapfullneighbors = FullNeighborList(cutoff, atoms)

    ase_neighbor_info = []
    for i in range(len(atoms), len(atoms_with_query)):
        indices, rel_positions, dist2 = asefullneighbors.get_neighbors(i, cutoff)
        # Filter out query atom neighbors
        rel_positions = rel_positions[indices < len(atoms)]
        dist2 = dist2[indices < len(atoms)]
        indices = indices[indices < len(atoms)]
        ase_neighbor_info.append((indices, rel_positions, dist2))

    asap_neighbor_info = asapfullneighbors.get_neighbors_querypoint(query_points)

    for asap_info, ase_info in zip(asap_neighbor_info, asap_neighbor_info):
        # Sort all lists based on neighbor distance (dist2)
        asap_sort = np.argsort(asap_info[2])
        ase_sort = np.argsort(ase_info[2])

        assert np.allclose(asap_info[2][asap_sort], ase_info[2][ase_sort]), "Distances equal"
        assert np.all(asap_info[0][asap_sort] == ase_info[0][ase_sort]), "Indices equal"
        assert np.all(asap_info[1][asap_sort] == ase_info[1][ase_sort]), "Differences equal"


@serial
def test_nbcelllocator_querypoint():
    # Setup test system
    element = "Cu"
    cutoff = 5
    atoms = FaceCenteredCubic(
        directions=[[1, 0, 0], [0, 1, 1], [0, 0, 1]],
        size=(9, 7, 5),
        symbol=element,
        debug=0,
    )

    # Test single query point
    query_points = np.array([[1.0, 2.0, 3.0]])
    compare_with_ase_neighborlist(atoms, query_points, cutoff)

    # Test multiple query points
    query_points = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    compare_with_ase_neighborlist(atoms, query_points, cutoff)

    # Test non-periodic boundaries
    query_points = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    atoms.set_pbc(False)
    compare_with_ase_neighborlist(atoms, query_points, cutoff)

    # Test non-periodic boundaries with query points outside boundaries
    query_points = np.array([[-1.0, 0.0, 0.0], [33.0, 21.0, 13.0]])
    atoms.set_pbc(False)
    compare_with_ase_neighborlist(atoms, query_points, cutoff)



from .atom import AtomMonteCarlo, AtomMonteCarloData
from .surface import SurfaceMonteCarlo, SurfaceMonteCarloData
#from atom_old import AtomMonteCarlo as AtomMonteCarloOld

def get_layers_from_size(size):
    n1 = int(round(pow(1.8 * size, 1./3.) / 2.))
    n2 = 2 * n1
    n3 = int(3./2. * n1)

    return [n1] * 6 + [n2] * 12 + [n3] * 8



import numpy as np
from asap3.test.nvt_module import main
from asap3.md.langevin import Langevin
from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

@pytest.mark.slow
@serial
def test_nvt_langevin():
    def lgvdynmaker(atoms, T0, dt, tau, rng, logint):
        # tau is the energy relaxation time.  The velocity relaxation time
        # should be the double.

        # Cannot pass the rng to Asap's Langevin, instead use it to create a seed
        seed = rng.integers(0, 1 << 30)
        print(f'Seed: {seed}')
        return Langevin(atoms, dt, temperature_K=T0, friction=1/(2*tau), logfile='-', loginterval=logint, seed=seed)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()
    main(lgvdynmaker, rng, sloppytime=True)

@pytest.mark.slow
@parallel
def test_par_nvt_langevin():
    def lgvdynmaker(atoms, T0, dt, tau, rng, logint):
        # tau is the energy relaxation time.  The velocity relaxation time
        # should be the double.

        # Cannot pass the rng to Asap's Langevin, instead use it to create a seed
        seed = rng.integers(0, 1 << 30)
        print(f'Seed: {seed}')
        return Langevin(atoms, dt, temperature_K=T0, friction=1/(2*tau), logfile='-', loginterval=logint, seed=seed)

    seed = np.random.SeedSequence(2718281828459045)
    seed = seed.spawn(world.size)[world.rank]
    rng = np.random.default_rng(seed)   # Use a fixed seed for reproducability

    main(lgvdynmaker, rng, parallel=True, sloppytime=True, failfluct=True)
    # Note: Fluctuation test always fail in parallel, as the boundaries
    # are free in order to test the migration better.  But that changes the
    # heat capacity, making the test fail.



"""Module for testing various NVT dynamics algorithms.

The actual tests import this module.
"""


from ase.units import fs, kB, GPa
import numpy as np
# matplotlib only imported in debugging mode (in function test_nvt)
# import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from asap3.md.velocitydistribution import (MaxwellBoltzmannDistribution,
                                             Stationary)
from asap3.md.bussi import Bussi
from asap3.md import MDLogger as _MDLogger
from asap3.test.pytest_markers import ReportTest
from asap3.mpi import world
try:
    from asap3 import MakeParallelAtoms
except ImportError:
    pass   # Probably installed as serial-only.
from asap3.test.nvt_module import make_atoms, MeasureEnergy, exponential


bulkmodulus = 140 * GPa    # Bulk modulus of Cu



class MDLogger(_MDLogger):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.stress:
            self.fmt = self.fmt[:-1] + '  '
            self.fmt += 9 * " %10.5f"
            self.fmt += '\n'
    def __call__(self):
        epot = self.atoms.get_potential_energy()
        ekin = self.atoms.get_kinetic_energy()
        temp = self.atoms.get_temperature()
        global_natoms = self.atoms.get_global_number_of_atoms()
        if self.peratom:
            epot /= global_natoms
            ekin /= global_natoms
        if self.dyn is not None:
            t = self.dyn.get_time() / (1000 * fs)
            dat = (t,)
        else:
            dat = ()
        dat += (epot + ekin, epot, ekin, temp)
        if self.stress:
            dat += tuple(self.atoms.get_stress(
                include_ideal_gas=True) / GPa)
            cell = np.ravel(self.atoms.cell.array)
            assert len(cell) == 9
            dat += tuple(cell)
        self.logfile.write(self.fmt % dat)
        self.logfile.flush()

class MeasureStuff(MeasureEnergy):
    "Observer measuring energy, temperature, volume and stress"
    def __init__(self, atoms):
        super().__init__(atoms)
        self.volumes = []
        self.stresses = []

    def __call__(self):
        super().__call__()
        self.volumes.append(self.atoms.get_volume())
        self.stresses.append(self.atoms.get_stress(include_ideal_gas=True))


def test_npt(atoms, nsteps, dt, T0, p0, dynmaker, initdyn=None, rng=None,
             intval=5, plot=False, sloppytime=False, failfluct=False,
             parallel=False):
    """Run NPT dynamics, testing the behaviour.

    First, thermalize the system at constant volume using a short
    time constant.  Then relax stress/pressure with a longer time, and 
    measure this relaxation time.  Finally, run the dynamics and measure
    fluctuations in enthalpy and volume.

    Currently, the test does not measure the relaxation time of the
    temperature.

    Parameters:
    atoms:      The atoms object.
    nsteps:     Length of simulation.
    dt:         Time step.
    T0:         Expected temperature.
    p0:         Expected pressure.
    dynmaker:   Function making the dynamics object.
    initdyn:    Function making initialization dynamics object, if different.
    rng:        Random number generator.
    intval:     Interval for taking data.
    plot:       Plot the swing-in temperature graph (default: False).
    sloppytime: Test of swing-in time is sloppy (+/- 50% instead of +/- 1%).
                (for nondeterministic dynamics)
    failfluct:  Test of energy fluctuations is expected to fail.
                (for dynamics that does not produce a true Canonical Ensemble)
    parallel:   Disable some tests in parallel test suite.
    """
    runtime = nsteps * dt
    taut = runtime / 20    # Energy relaxation time in ideal gas
    taup = runtime / 10

    # Initial relaxation of temperature is done using Bussi dynamics
    dyn = Bussi(atoms, dt, temperature_K=T0, taut=taut, rng=rng)
    dyn.attach(MDLogger(dyn, atoms, '-', peratom=True, stress=True), interval=nsteps//10)
    dyn.run(nsteps)
    
    # We pass *half* the energy relaxation time to the generator of the dynamics,
    # as we have a solid, where the relaxation time will be twice that of the
    # ideal gas, since the same amount of potential and kinetic energy needs
    # to be added to the system.
    if initdyn is None:
        initdyn = dynmaker   # Normally use the same dynamics for both tests
    dyn = initdyn(atoms, T0, p0, dt, taut=taut/2, taup=taup, rng=rng, logint=1000)
    measure = MeasureStuff(atoms)
    dyn.attach(measure, interval=1)
    dyn.run(nsteps)
    stresses = np.array(measure.stresses)
    pressures = -stresses[:,:3].sum(axis=1) / 3

    # Fit temperature curve
    pressures = pressures[5:]
    times = np.arange(len(pressures)) * dt
    if not parallel:
        try:
            (DeltaPfit, taup_fit, P_fit), _ = curve_fit(exponential, times, pressures, (-p0, taup, p0))
        except RuntimeError:
            # Fitting failed
            DeltaPfit, taup_fit, P_fit = 0, 1, 0
            
    # Run again with smaller tau
    tausmall = runtime / 100
    logint = 5000
    if failfluct:
        logint = 1000
    dyn = dynmaker(atoms, T0, p0, dt, taut=tausmall, taup=tausmall, rng=rng, logint=logint)
    measure = MeasureStuff(atoms)
    dyn.attach(measure, interval=intval)
    com_before = atoms.get_center_of_mass()
    if parallel:
        print(f'Atoms on rank {world.rank}: {len(atoms)}')
    if failfluct:
        # No need for good statistics if it fails anyway
        dyn.run(nsteps)
    else:
        dyn.run(nsteps*10)
    if parallel:
        print(f'Atoms on rank {world.rank}: {len(atoms)}')
    com_after = atoms.get_center_of_mass()
    energies = np.array(measure.energies)
    stresses = np.array(measure.stresses)
    pressures2 = -stresses[:,:3].sum(axis=1) / 3
    volumes = np.array(measure.volumes)
    enthalpies = energies + p0 * volumes
    enthalpies = enthalpies[len(energies)//10:]
    temperatures2 = measure.temperatures
    temperatures2 = temperatures2[len(temperatures2)//5:]
    pressures2 = pressures2[len(pressures2)//5:]
    stresses = stresses[len(stresses)//5:]
    times2 = np.arange(len(temperatures2)) * dt * intval

    stdH = np.std(enthalpies)
    stdV = np.std(volumes)
    avgV = np.mean(volumes)
    avgT = np.mean(temperatures2)
    avgP = np.mean(pressures2)
    avgstr = np.mean(stresses, axis=0)
    # Expected enthalpy fluctuation: sqrt(k_B T^2 3 N k_B) = k_B * T * sqrt(3 * N)
    expected = kB * T0 * np.sqrt(3 * len(atoms))
    expectedV = np.sqrt(kB * T0 * avgV / bulkmodulus)

    # Output results
    if not parallel:
        print(f"Part 1 pressure:   {P_fit/GPa:.2f} GPa    (expected {p0/GPa:.2f})")
        print(f"Time constant:     {taup_fit/fs:.1f} fs  (expected {taup/fs:.1f}  error {(taup_fit / taup - 1) * 100:.1f}%)")
        print(f"Initial pressure offset:  {DeltaPfit/GPa:.2f} GPa")
        taup_error = 0.5 * taup/fs   # Larger than for NVT as thermostat and barostat interact
        ReportTest('Time constant', taup_fit/fs, taup/fs, taup_error)

    print(f"Observed enthalpy fluctuation: {stdH:.2f} eV")
    print(f"Expected enthalpy fluctuation: {expected:.2f} eV")
    print(f"Error: {(stdH / expected - 1) * 100:.1f}%")
    
    print(f"Average volume {avgV:.2f} Å^3")
    print(f"Observed volume fluctuation: {stdV:.2f} Å^3")
    print(f"Expected volume fluctuation: {expectedV:.2f} Å^3")
    print(f"Error: {(stdV / expectedV - 1) * 100:.1f}%")
  
    if failfluct:
        print('Failed fluctuations EXPECTED for this dynamics!')
    else:
        ReportTest("Energy fluctuations", stdH, expected, 0.15*expected)
        ReportTest("Volume fluctuations", stdV, expectedV, 0.15*expectedV)

    # Temperature error: We should be able to detect a error of 1/N_atoms
    maxtemperr = 2/3 * 3/atoms.get_number_of_degrees_of_freedom()
    # ... but not if we don't have good statistics.
    if parallel:
        maxtemperr *= 20  # System is much larger and run is shorter
    elif failfluct:
        maxtemperr *= 4
    maxperr = 0.03

    print(f'Observed average temperature:  {avgT:.2f} K   (expected {T0:.2f} K)')
    print(f'Error: {(avgT / T0 - 1) * 100:.1f}%  (max: {maxtemperr * 100:.1f}%)')
    ReportTest('Temperature', avgT, T0, T0 * maxtemperr)

    obstress = ' '.join([f'{s/GPa:.3f}' for s in avgstr])
    print(f'Observed average stress [{obstress}] GPa')
    print(f'Observed average pressure:  {avgP/GPa:.3f} GPa   (expected {p0/GPa:.3f} GPa)')
    print(f'Error: {(avgP / p0 - 1) * 100:.3f}%  (max: {maxperr * 100:.1f}%)')
    ReportTest('Pressure', avgP, p0, p0 * maxperr)

    print('Center of mass before:', com_before)
    print('Center of mass after: ', com_after)

    if plot:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.plot(times, pressures, 'b.')
        if not parallel:
            ax.plot(times, exponential(times, DeltaPfit, taup_fit, P_fit), 'k-')
        ax.plot([times[0], times[-1]], [p0, p0], 'k:')
        ax.set_title('Pressure during initial phase')
        fig2, ax2 = plt.subplots()
        ax2.plot(times2, temperatures2, 'b.')
        ax2.set_title('Temperature')
        fig3, ax3 = plt.subplots()
        ax3.plot(times2, pressures2, 'b.')
        ax3.plot([times2[0], times2[-1]], [p0, p0], 'k:')
        ax3.set_title('Pressure')
        plt.show(block=True)


def main(dynmaker, rng, repeat=1, plot=False, sloppytime=False, failfluct=False,
         initdyn=None, parallel=False):
    T0 = 300
    p0 = 2.0 * GPa
    dt = 5 * fs
    nsteps = 20000

    if parallel:
        T0 = 1000
        nsteps //= 4  # The system is much bigger
        # Plot only on master
        if plot:
            plot = world.rank == 0

    for i in range(repeat):
        atoms = make_atoms(T0 / 10, rng, parallel=parallel)
        c = atoms.cell 
        c[0,0] *= 0.998
        c[1,1] *= 1.002
        atoms.cell = c
        test_npt(atoms, nsteps, dt, T0, p0, dynmaker=dynmaker, initdyn=initdyn,
                 rng=rng, plot=plot, sloppytime=sloppytime, failfluct=failfluct,
                 parallel=parallel)
    if parallel:
        world.barrier()  # Needed if plotting, at least.

if __name__ == "__main__":
    print("This it not a test, but a module imported from a few tests.")



"""
Tests that polyhedral template matching works on the simplest possible system.
"""

from ase.lattice.cubic import FaceCenteredCubic, BodyCenteredCubic
from ase.lattice.hexagonal import HexagonalClosedPacked
from ase.build import bulk
from asap3.analysis import PTM
from asap3 import MakeParallelAtoms
from asap3.mpi import world
import numpy as np
from asap3.test.pytest_markers import ReportTest

import pytest



def distribute(atoms, layout):
    if world.rank != 0:
        atoms = None
    if world.size > 1:
        atoms = MakeParallelAtoms(atoms, layout)
    return atoms

def mysum(x):
    """Sum arrays across processors, taking into account that the length may vary.

    No array is longer than 10, so we just extend them all.
    """
    data = np.zeros(10, x.dtype)
    data[:len(x)] = x
    world.sum(data)
    return data

@pytest.mark.core
@pytest.mark.skipif(world.size not in (1, 2, 4, 8), reason='Number of cores must be 1,2,4 or 8')
def test_ptm(cpulayout):
    atoms = distribute(FaceCenteredCubic("Cu", size=(14,14,14), pbc=False), cpulayout)

    data = PTM(atoms, rmsd_max=0.1, calculate_strains=True)
    print("DONE")
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of FCC atoms", cnt[1], 13*13*13*4, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 13*13*13*4, 0)

    data = PTM(atoms, rmsd_max=0.1, calculate_strains=True, quick=True)
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of FCC atoms", cnt[1], 13*13*13*4, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 13*13*13*4, 0)

    data = PTM(atoms, rmsd_max=0.1, target_structures=('fcc', 'hcp'))
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of FCC atoms", cnt[1], 13*13*13*4, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 13*13*13*4, 0)

    atoms = distribute(BodyCenteredCubic("Fe", size=(14,14,14), pbc=(True,True,False)), cpulayout)
    data = PTM(atoms, rmsd_max=0.001, cutoff=8.0)
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of BCC atoms", cnt[3], 14*14*12*2, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 14*14*12*2, 0)

    atoms = distribute(HexagonalClosedPacked("Mg", size=(14,14,14), pbc=(True,True,False)), cpulayout)
    cutoff = atoms.get_cell()[0,0] / 4 / np.sqrt(3) * 0.95
    print("Cutoff", cutoff)
    data = PTM(atoms, rmsd_max=0.01, cutoff=cutoff)  # Need a slightly larger rmsd as c/a deviates from ideal.
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of HCP atoms", cnt[2], 14*14*13*2, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 14*14*13*2, 0)
    #atoms.set_tags(data['structure'])
    #view(atoms)

    atoms = bulk("Mg", orthorhombic=True)
    atoms = atoms.repeat((14,14,14))
    atoms.set_pbc((True, True, False))
    atoms = distribute(atoms, cpulayout)
    data = PTM(atoms, rmsd_max=0.01)  # Need a slightly larger rmsd as c/a deviates from ideal.
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of HCP atoms", cnt[2], 14*14*13*4, 0)
    ReportTest("Number of non-classified atoms", cnt[0], atoms.get_global_number_of_atoms() - 14*14*13*4, 0)

    # Make a combined system

    atoms = bulk("Mg", orthorhombic=True)
    atoms = atoms.repeat((14,8,14))

    atoms2 = bulk("Fe", orthorhombic=True).repeat((6,6,6))
    uc1 = atoms.get_cell()
    uc2 = atoms2.get_cell()
    factor1 = uc1[0,0] / uc2[0,0]
    factor2 = uc1[1,1] / uc2[1,1]
    uc2 = uc2 * np.array((factor1, factor2, 0.5*(factor1+factor2)))
    atoms2.set_cell(uc2, scale_atoms=True)
    # Combine unit cell, stack BCC atoms on top of HCP atoms.
    uc3 = np.array(uc1)
    dz = uc3[2,2]
    uc3[2,2] += uc2[2,2]
    atoms.set_cell(uc3, scale_atoms=False)
    atoms2.set_positions(atoms2.get_positions() + np.array((0.1,0.1,dz))[np.newaxis,:])
    atoms.extend(atoms2)
    atoms = distribute(atoms, cpulayout)

    data = PTM(atoms, rmsd_max=0.005)
    cnt = np.bincount(data['structure'])
    cnt = mysum(cnt)
    print(cnt)
    print(data['info'])

    ReportTest("Number of HCP atoms", cnt[2], 14*8*13*4, 0)
    ReportTest("Number of BCC atoms", cnt[3], 6*6*4*2, 0)
    ReportTest("Number of FCC atoms", cnt[1], 0, 0)
    ReportTest("All classes of atoms", cnt[:4].sum(), atoms.get_global_number_of_atoms(), 0)



"""Module for testing various NVT dynamics algorithms.

The actual tests import this module.
"""


from ase.units import fs, kB
from ase.build import bulk
import asap3
import numpy as np
# matplotlib only imported in debugging mode (in function test_nvt)
# import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from asap3.md.velocitydistribution import (MaxwellBoltzmannDistribution,
                                             Stationary)
from asap3.test.pytest_markers import ReportTest
from asap3.io import Trajectory
from asap3.mpi import world
try:
    from asap3 import MakeParallelAtoms
except ImportError:
    pass   # Probably installed as serial-only.

def make_atoms(T, rng, size=(2,2,2), parallel=False):
    "Make atomic system suitable for test."
    atoms = bulk('Cu', cubic=True)
    atoms *= size
    if parallel:
        if world.rank == 0:
            atoms *= (2, 2, 2 * world.size)
            atoms.set_pbc(False)
            atoms.center(vacuum=2.0, axis=(0, 1))
            atoms.set_pbc(True)
        else:
            atoms = None
        atoms = MakeParallelAtoms(atoms, (1, 1, world.size))
    MaxwellBoltzmannDistribution(atoms, temperature_K=T, rng=rng)
    Stationary(atoms)
    atoms.calc = asap3.EMT()
    return atoms

class MeasureEnergy:
    "Observer measuring the energy and temperature."
    def __init__(self, atoms):
        self.atoms = atoms
        self.energies = []
        self.temperatures = []

    def __call__(self):
        e = self.atoms.get_potential_energy() + self.atoms.get_kinetic_energy()
        T = self.atoms.get_temperature()
        self.energies.append(e)
        self.temperatures.append(T)


def exponential(t, A, tau, B):
    "We fit stuff to exponentials"
    return A * np.exp(-t / tau) + B

def test_nvt(atoms, nsteps, dt, T0, dynmaker, initdyn=None, rng=None,
             intval=5, plot=False, sloppytime=False, failfluct=False,
             parallel=False, trajectory=None):
    """Run NVT dynamics, testing the behaviour.

    Parameters:
    atoms:      The atoms object.
    nsteps:     Length of simulation.
    dt:         Time step.
    T0:         Expected temperature
    dynmaker:   Function making the dynamics object.
    initdyn:    Function making initialization dynamics object, if different.
    rng:        Random number generator.
    intval:     Interval for taking data.
    plot:       Plot the swing-in temperature graph (default: False).
    sloppytime: Test of swing-in time is sloppy (+/- 50% instead of +/- 1%).
                (for nondeterministic dynamics)
    failfluct:  Test of energy fluctuations is expected to fail.
                (for dynamics that does not produce a true Canonical Ensemble)
    parallel:   Disable some tests in parallel test suite.
    """
    runtime = nsteps * dt
    tau = runtime / 10    # Energy relaxation time in ideal gas

    # We pass *half* the energy relaxation time to the generator of the dynamics,
    # as we have a solid, where the relaxation time will be twice that of the
    # ideal gas, since the same amount of potential and kinetic energy needs
    # to be added to the system.
    if initdyn is None:
        initdyn = dynmaker   # Normally use the same dynamics for both tests
    dyn = initdyn(atoms, T0, dt, tau/2, rng=rng, logint=1000)
    measure = MeasureEnergy(atoms)
    dyn.attach(measure, interval=1)
    dyn.run(nsteps)
    #energies = measure.energies
    temperatures = measure.temperatures

    # Fit temperature curve
    temperatures = temperatures[5:]
    times = np.arange(len(temperatures)) * dt
    if not parallel:
        (DeltaTfit, tau_fit, T_fit), _ = curve_fit(exponential, times, temperatures, (-T0, tau, T0))

    # Run again with smaller tau
    tausmall = runtime / 100
    dyn = dynmaker(atoms, T0, dt, tausmall, rng=rng, logint=5000)
    measure = MeasureEnergy(atoms)
    dyn.attach(measure, interval=intval)
    com_before = atoms.get_center_of_mass()
    if trajectory:
        traj = Trajectory(trajectory, "w", atoms)
        dyn.attach(traj, interval=nsteps//5)
    if failfluct:
        # No need for good statistics if it fails anyway
        dyn.run(nsteps)
    else:
        dyn.run(nsteps*10)
    com_after = atoms.get_center_of_mass()
    energies = measure.energies
    energies = energies[len(energies)//10:]
    temperatures2 = measure.temperatures
    temperatures2 = temperatures2[len(temperatures2)//5:]
    times2 = np.arange(len(temperatures2)) * dt * intval

    stdE = np.std(energies)
    avgT = np.mean(temperatures2)
    # Expected energy fluctuation: sqrt(k_B T^2 3 N k_B) = k_B * T * sqrt(3 * N)
    expected = kB * T0 * np.sqrt(3 * len(atoms))

    # Output results
    if not parallel:
        print(f"Part 1 temperature:   {T_fit:.2f} K    (expected {T0:.2f})")
        print(f"Time constant:       {tau_fit/fs:.1f} fs  (expected {tau/fs:.1f}  error {(tau_fit / tau - 1) * 100:.1f}%)")
        print(f"Initial temperature offset:  {DeltaTfit:.2f} K")
        if sloppytime:
            tau_error = 0.5 * tau/fs
        else:
            tau_error = 0.01 * tau/fs
        ReportTest('Time constant', tau_fit/fs, tau/fs, tau_error)

    print(f"Observed energy fluctuation: {stdE:.2f} eV")
    print(f"Expected energy fluctuation: {expected:.2f} eV")
    print(f"Error: {(stdE / expected - 1) * 100:.1f}%")
    if failfluct:
        print('Failed fluctuations EXPECTED for this dynamics!')
    else:
        ReportTest("Energy fluctuations", stdE, expected, 0.15*expected)

    # Temperature error: We should be able to detect a error of 1/N_atoms
    maxtemperr = 2/3 * 3/atoms.get_number_of_degrees_of_freedom()
    # ... but not if we don't have good statistics.
    if parallel:
        maxtemperr *= 20  # System is much larger and run is shorter
    elif failfluct:
        maxtemperr *= 4

    print(f'Observed average temperature:  {avgT:.2f} K   (expected {T0:.2f} K)')
    print(f'Error: {(avgT / T0 - 1) * 100:.1f}%  (max: {maxtemperr * 100:.1f}%)')
    ReportTest('Temperature', avgT, T0, T0 * maxtemperr)

    print('Center of mass before:', com_before)
    print('Center of mass after: ', com_after)

    if plot:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.plot(times, temperatures, 'b.')
        if not parallel:
            ax.plot(times, exponential(times, DeltaTfit, tau_fit, T_fit), 'k-')
        fig2, ax2 = plt.subplots()
        ax2.plot(times2, temperatures2, 'b.')
        plt.show(block=True)


def main(dynmaker, rng, repeat=1, plot=False, sloppytime=False, failfluct=False,
         initdyn=None, parallel=False, trajectory=None):
    T0 = 300
    dt = 5 * fs
    nsteps = 20000

    if parallel:
        T0 = 1000
        nsteps //= 4  # The system is much bigger
        # Plot only on master
        if plot:
            plot = world.rank == 0

    for i in range(repeat):
        atoms = make_atoms(T0 / 10, rng, parallel=parallel)
        if parallel:
            print(f'Atoms on rank {world.rank}: {len(atoms)}')
        test_nvt(atoms, nsteps, dt, T0, dynmaker=dynmaker, initdyn=initdyn,
                 rng=rng, plot=plot, sloppytime=sloppytime, failfluct=failfluct,
                 parallel=parallel, trajectory=trajectory)
        if parallel:
            print(f'Atoms on rank {world.rank}: {len(atoms)}')
    if parallel:
        world.barrier()  # Needed if plotting, at least.

if __name__ == "__main__":
    print("This it not a test, but a module imported from a few tests.")



import numpy as np
from asap3 import CoordinationNumbers
from ase import units
from .langmuirExpression import getCoverages,getEnergies


class adscalc:
	
	
	def __init__(self,calc,pressure=1E2,temperature=300,species = "AuCO", relax_calc=None):
		self.calc = calc
		self.relax_calc = relax_calc
		self.species = species
		self.pressure = float(pressure)
		self.temperature = float(temperature)
		
		self.covbyc = np.array(getCoverages(T=temperature,P=pressure,species = self.species)) #Once and for all since it is not system specific.
		
		self.ebyc=np.array(getEnergies(self.species))
		
		
		self.active = True
		
	def set_atoms(self,atoms):
		self.calc.set_atoms(atoms)
		
	
	
	def get_potential_energy(self,atoms):
		e = self.calc.get_potential_energy(atoms)
		if self.active:
			acoord = CoordinationNumbers(atoms) #Contains coordination indexed by atoms index.
			
			eret = self.ebyc[acoord]*self.covbyc[acoord]
			
			#Debug:
			#print "Coordinations:"+str(acoord)
			#print "return array:"+str(eret)
			e+= eret.sum()
		return e
		
	def use_adsorbates(self, active):
		self.active = active
		
	#Overrides needed to write to pickle trajectory!
	def get_forces(self,atmref):
		return self.calc.get_forces(atmref)
		
	def get_stress(self,atmref):
		return self.calc.get_stress(atmref)
		
	def get_neighborlist(self):
		return self.calc.get_neighborlist()

	def calculation_required(self, *args, **kwargs):
		return self.calc.calculation_required(*args, **kwargs)



# Copied from GPAW (https://trac.fysik.dtu.dk/projects/gpaw/)

import time
import math
import sys

from ase.parallel import world

def function_timer(func, *args, **kwargs):
    out = kwargs.pop('timeout', sys.stdout)
    t1 = time.time()
    r = func(*args, **kwargs)
    t2 = time.time()
    print(t2 - t1, file=out)
    return r

class Timer:
    def __init__(self, print_levels=1000):
        self.timers = {}
        self.t0 = time.time()
        self.running = []
        self.print_levels = print_levels
        
    def start(self, name):
        names = tuple(self.running + [name])
        self.timers[names] = self.timers.get(names, 0.0) - time.time()
        self.running.append(name)
        
    def stop(self, name=None):
        if name is None: name = self.running[-1]
        names = tuple(self.running)
        running = self.running.pop()
        if name != running:
            raise RuntimeError('Must stop timers by stack order.  '
                               'Requested stopping of %s but topmost is %s'
                               % (name, running))
        self.timers[names] += time.time()
            
    def get_time(self, *names):
#        print self.timers, names
        return self.timers[names]
                
    def write(self, out=sys.stdout):
        while self.running:
            self.stop()
        if len(self.timers) == 0:
            return

        t0 = time.time()
        tot = t0 - self.t0

        n = max([len(names[-1]) + len(names) for names in self.timers]) + 1
        out.write('\n%-*s    incl.     excl.\n' % (n, 'Timing:'))
        out.write('%s\n' % ('=' * 60))
        tother = tot
        
        inclusive = self.timers.copy()

        exclusive = self.timers
        keys = exclusive.keys()
        keys.sort()
        for names in keys:
            t = exclusive[names]
            if len(names) > 1:
                if len(names) < self.print_levels + 1:
                    exclusive[names[:-1]] -= t
            else:
                tother -= t
        exclusive[('Other',)] = tother
        inclusive[('Other',)] = tother
        keys.append(('Other',))
        for names in keys:
            t = exclusive[names]
            tinclusive = inclusive[names]
            r = t / tot
            p = 100 * r
            i = int(40 * r + 0.5)
            if i == 0:
                bar = '|'
            else:
                bar = '|%s|' % ('-' * (i - 1))
            level = len(names)
            if level > self.print_levels:
                continue
            name = (level - 1) * ' ' + names[-1] + ':'
            out.write('%-*s%9.3f %9.3f %5.1f%% %s\n' %
                      (n, name, tinclusive, t, p, bar))
        out.write('%s\n' % ('=' * 60))
        out.write('%-*s%9.3f %5.1f%%\n' % (n + 10, 'Total:', tot, 100.0))
        out.write('%s\n' % ('=' * 60))
        out.write('date: %s\n' % time.asctime())

    def add(self, timer):
        for name, t in timer.timers.items():
            self.timers[name] = self.timers.get(name, 0.0) + t


class DebugTimer(Timer):
    def __init__(self, print_levels=1000, comm=world, txt=sys.stdout):
        Timer.__init__(self, print_levels)
        ndigits = 1 + int(math.log10(comm.size))
        self.srank = '%0*d' % (ndigits, comm.rank)
        self.txt = txt

    def start(self, name):
        Timer.start(self, name)
        t = self.timers[tuple(self.running)] + time.time()
        self.txt.write(f'T{self.srank} >> {name} ({t:7.5f}s) started\n')

    def stop(self, name=None):
        if name is None: name = self.running[-1]
        t = self.timers[tuple(self.running)] + time.time()
        self.txt.write(f'T{self.srank} << {name} ({t:7.5f}s) stopped\n')
        Timer.stop(self, name)


class StepTimer(Timer):
    """Step timer to print out timing used in computation steps.
    
    Use it like this::

      from gpaw.utilities.timing import StepTimer
      st = StepTimer()
      ...
      st.write_now('step 1')
      ...
      st.write_now('step 2')

    The parameter write_as_master_only can be used to force the timer to
    print from processess that are not the mpi master process.
    """
    
    def __init__(self, out=sys.stdout, name=None, write_as_master_only=True):
        Timer.__init__(self)
        if name is None:
            name = '<%s>' % sys._getframe(1).f_code.co_name
        self.name = name
        self.out = out
        self.alwaysprint = not write_as_master_only
        self.now = 'temporary now'
        self.start(self.now)


    def write_now(self, mark=''):
        self.stop(self.now)
        if self.alwaysprint or world.rank == MASTER:
            print(self.name, mark, self.gettime(self.now), file=self.out)
        self.out.flush()
        del self.timers[self.now]
        self.start(self.now)





import numpy as np

from asap3.MonteCarlo.Base import GlobalOptimizer
from ase.units import kB
from ase.optimize import LBFGS

class BasinHopping(GlobalOptimizer):
    def __init__(self, atoms, log='-', traj=None, localopt=LBFGS, fmax=1e-2):
        GlobalOptimizer.__init__(self, atoms, log, traj)
        self.localoptimizer = localopt
        self.fmax = fmax
        self.Eo = atoms.get_potential_energy()
        self.En = 1.0e32
        self.P = 0.0

        self.log_header = ('%-15s  %-12s  %-6s  %-7s\n' %
                           ('Energy', 'Probability', 'Accept', 'OptStep'))

    def initialize(self, temp=1000.0):
        GlobalOptimizer.initialize(self)
        self.temp = temp
        self.log_info = '\nTemperature: %i\n' % (temp,)

    def evaluate_move(self):
        dyn = self.localoptimizer(self.atoms, logfile=None)
        dyn.run(fmax=self.fmax, steps=1000)

        self.En = self.atoms.get_potential_energy()
        self.P = min(1.0, np.exp((self.Eo - self.En) / (self.temp * kB)))
        accept = self.P > np.random.uniform()

        self.log_string = ('%-15.6f  %-12.6f  %-6s  %-7i\n' %
                           (self.En, self.P, accept, dyn.nsteps))

        return accept

    def accept_move(self):
        self.Eo = self.En
        GlobalOptimizer.accept_move(self)




"""Check that energy is correct even after wrapping through periodic boundary conditions.
"""

from ase.lattice.cubic import FaceCenteredCubic
from asap3 import EMT
from numpy.random import default_rng
from asap3.test.pytest_markers import ReportTest, serial

import pytest

@serial
@pytest.mark.core
def test_pbc_wrap():
    random = default_rng()

    ref_atoms = FaceCenteredCubic(size=(7,7,7), symbol="Cu", pbc=(True, False, True))
    ref_atoms.calc = EMT()

    ref_energy = ref_atoms.get_potential_energy()
    ref_energies = ref_atoms.get_potential_energies()
    ref_forces = ref_atoms.get_forces()

    passes = 5
    for ps in range(passes):
        print("Pass", ps, "of", passes)

        atoms = ref_atoms.copy()
        atoms.calc = EMT()
        nat = random.integers(0, len(atoms))
        assert nat < len(atoms)
        pos0 = atoms[nat].position
        cell = atoms.get_cell()
        for d in range(1,4):
            for dx in (-d, 0, d):
                #for dy in (-d, 0, d):
                for dy in (0,):
                    for dz in (-d, 0 ,d):
                        deltar = dx * cell[0] + dy * cell[1] + dz * cell[2]
                        atoms[nat].position = pos0 + deltar
                        label = "(%2d, %2d, %2d)" % (dx, dy, dz)
                        ReportTest("Pot. energy   "+label,
                                atoms.get_potential_energy(), ref_energy, 1e-6)
                        de = (atoms.get_potential_energies() - ref_energies)
                        ReportTest("Pot. energies "+label, de.max(), 0.0, 1e-6)
                        df = (atoms.get_forces() - ref_forces)
                        ReportTest("Forces        "+label, df.max(), 0.0, 1e-6)
                    




"""Check that ASE.calculators.emt and asap3.EMT give the same."""

from ase.lattice.compounds import *
from ase.lattice.cubic import *
from ase.calculators.emt import EMT as EMT_ASE
from asap3 import EMT as EMT_ASAP
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

@serial
@pytest.mark.core
def test_ase_emt():
    elements = ("Ni", "Cu", "Pd", "Ag", "Pt")
    for e1 in elements:
        for e2 in elements:
            atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(1,1,2),
                                    symbol=e1, pbc=(1,0,1), debug=0)
            atoms.calc = EMT_ASE(asap_cutoff=True)
            e_e1_ase = atoms.get_potential_energy()
            atoms.calc = EMT_ASAP()
            e_e1_asap = atoms.get_potential_energy()
            natoms = len(atoms)

            print(f"{e1} energy (ASE) \t{e_e1_ase/natoms:.5f}")
            print(f"{e1} energy (ASAP)\t{e_e1_asap/natoms:.5f}")

            atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(1,1,2),
                                    symbol=e2, pbc=(1,0,1), debug=0)
            atoms.calc = EMT_ASE(asap_cutoff=True)
            e_e2_ase = atoms.get_potential_energy()
            atoms.calc = EMT_ASAP()
            e_e2_asap = atoms.get_potential_energy()

            print(f"{e2} energy (ASE) \t{e_e2_ase/natoms:.5f}")
            print(f"{e2} energy (ASAP)\t{e_e2_asap/natoms:.5f}")

            atoms = L1_2(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(1,1,2),
                        symbol=(e1, e2), latticeconstant=3.95, pbc=(1,0,1), 
                        debug=0)
            
            atoms.calc = EMT_ASE(asap_cutoff=True)
            e_alloy_ase = atoms.get_potential_energy() - (2*e_e1_ase + 6*e_e2_ase)/8
            atoms.calc = EMT_ASAP()
            e_alloy_asap = atoms.get_potential_energy() - (2*e_e1_asap + 6*e_e2_asap)/8

            print(f"Alloy energy (ASE) \t{e_alloy_ase/natoms:.5f}")
            print(f"Alloy energy (ASAP)\t{e_alloy_asap/natoms:.5f}")
            ReportTest(f"{e1}{e2}_3 alloy energy", e_alloy_ase, e_alloy_asap, 1e-4)


